"""Analysis results and their aggregation --- :mod:`MDAnalysis.analysis.results`
================================================================================

Module introduces two classes, :class:`Results` and :class:`ResultsGroup`,
used for storing and aggregating data in
:meth:`MDAnalysis.analysis.base.AnalysisBase.run()`, respectively.


Classes
-------

The :class:`Results` class is an extension of a built-in dictionary
type, that holds all assigned attributes in :attr:`self.data` and 
allows for access either via dict-like syntax, or via class-like syntax:

.. code-block:: python

    from MDAnalysis.analysis.results import Results
    r = Results()
    r.array = [1, 2, 3, 4]
    assert r['array'] == r.array == [1, 2, 3, 4]


The :class:`ResultsGroup` can merge multiple :class:`Results` objects.
It is mainly used by :class:`MDAnalysis.analysis.base.AnalysisBase` class, 
that uses :meth:`ResultsGroup.merge()` method to aggregate results from
multiple workers, initialized during a parallel run:

.. code-block:: python

    from MDAnalysis.analysis.results import Results, ResultsGroup
    import numpy as np
    
    r1, r2 = Results(), Results()
    r1.masses = [1, 2, 3, 4, 5]
    r2.masses = [0, 0, 0, 0]
    r1.vectors = np.arange(10).reshape(5, 2)
    r2.vectors = np.arange(8).reshape(4, 2)

    group = ResultsGroup(
        lookup = {
            'masses': ResultsGroup.flatten_sequence,
            'vectors': ResultsGroup.ndarray_vstack
            }
        )

    r = group.merge([r1, r2])
    assert r.masses == list((*r1.masses, *r2.masses))
    assert (r.vectors == np.vstack([r1.vectors, r2.vectors])).all()
"""

from collections import UserDict
import numpy as np
from typing import Callable, Sequence


class Results(UserDict):
    r"""Container object for storing results.

    :class:`Results` are dictionaries that provide two ways by which values
    can be accessed: by dictionary key ``results["value_key"]`` or by object
    attribute, ``results.value_key``. :class:`Results` stores all results
    obtained from an analysis after calling :meth:`~AnalysisBase.run()`.

    The implementation is similar to the :class:`sklearn.utils.Bunch`
    class in `scikit-learn`_.

    .. _`scikit-learn`: https://scikit-learn.org/
    .. _`sklearn.utils.Bunch`: https://scikit-learn.org/stable/modules/generated/sklearn.utils.Bunch.html

    Raises
    ------
    AttributeError
        If an assigned attribute has the same name as a default attribute.

    ValueError
        If a key is not of type ``str`` and therefore is not able to be
        accessed by attribute.

    Examples
    --------
    >>> from MDAnalysis.analysis.base import Results
    >>> results = Results(a=1, b=2)
    >>> results['b']
    2
    >>> results.b
    2
    >>> results.a = 3
    >>> results['a']
    3
    >>> results.c = [1, 2, 3, 4]
    >>> results['c']
    [1, 2, 3, 4]


    .. versionadded:: 2.0.0

    .. versionchanged:: 2.8.0
        Moved :class:`Results` to :mod:`MDAnalysis.analysis.results`
    """

    def _validate_key(self, key):
        if key in dir(self):
            raise AttributeError(
                f"'{key}' is a protected dictionary attribute"
            )
        elif isinstance(key, str) and not key.isidentifier():
            raise ValueError(f"'{key}' is not a valid attribute")

    def __init__(self, *args, **kwargs):
        kwargs = dict(*args, **kwargs)
        if "data" in kwargs.keys():
            raise AttributeError(f"'data' is a protected dictionary attribute")
        self.__dict__["data"] = {}
        self.update(kwargs)

    def __setitem__(self, key, item):
        self._validate_key(key)
        super().__setitem__(key, item)

    def __setattr__(self, attr, val):
        if attr == "data":
            super().__setattr__(attr, val)
        else:
            self.__setitem__(attr, val)

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as err:
            raise AttributeError(
                f"'Results' object has no attribute '{attr}'"
            ) from err

    def __delattr__(self, attr):
        try:
            del self[attr]
        except KeyError as err:
            raise AttributeError(
                f"'Results' object has no attribute '{attr}'"
            ) from err

    def __getstate__(self):
        return self.data

    def __setstate__(self, state):
        self.data = state


class ResultsGroup:
    """
    Management and aggregation of results stored in :class:`Results` instances.

    A :class:`ResultsGroup` is an optional description for :class:`Result` "dictionaries"
    that are used in analysis classes based on :class:`AnalysisBase`. For each *key* in a
    :class:`Result` it describes how multiple pieces of the data held under the key are
    to be aggregated. This approach is necessary when parts of a trajectory are analyzed
    independently (e.g., in parallel) and then need to me merged (with :meth:`merge`) to
    obtain a complete data set.

    Parameters
    ----------
    lookup : dict[str, Callable], optional
        aggregation functions lookup dict, by default None

    Examples
    --------

    .. code-block:: python

        from MDAnalysis.analysis.results import ResultsGroup, Results
        group = ResultsGroup(lookup={'mass': ResultsGroup.float_mean})
        obj1 = Results(mass=1)
        obj2 = Results(mass=3)
        assert {'mass': 2.0} == group.merge([obj1, obj2])


    .. code-block:: python

        # you can also set `None` for those attributes that you want to skip
        lookup = {'mass': ResultsGroup.float_mean, 'trajectory': None}
        group = ResultsGroup(lookup)
        objects = [Results(mass=1, skip=None), Results(mass=3, skip=object)]
        assert group.merge(objects, require_all_aggregators=False) == {'mass': 2.0}

    .. versionadded:: 2.8.0
    """

    def __init__(self, lookup: dict[str, Callable] = None):
        self._lookup = lookup

    def merge(
        self, objects: Sequence[Results], require_all_aggregators: bool = True
    ) -> Results:
        """Merge multiple Results into a single Results instance.

        Merge multiple :class:`Results` instances into a single one, using the
        `lookup` dictionary to determine the appropriate aggregator functions for
        each named results attribute. If the resulting object only contains a single
        element, it just returns it without using any aggregators.

        Parameters
        ----------
        objects : Sequence[Results]
            Multiple :class:`Results` instances with the same data attributes.
        require_all_aggregators : bool, optional
            if True, raise an exception when no aggregation function for a
            particular argument is found. Allows to skip aggregation for the
            parameters that aren't needed in the final object --
            see :class:`ResultsGroup`.

        Returns
        -------
        Results
            merged :class:`Results`

        Raises
        ------
        ValueError
            if no aggregation function for a key is found and ``require_all_aggregators=True``
        """
        if len(objects) == 1:
            merged_results = objects[0]
            return merged_results

        merged_results = Results()
        for key in objects[0].keys():
            agg_function = self._lookup.get(key, None)
            if agg_function is not None:
                results_of_t = [obj[key] for obj in objects]
                merged_results[key] = agg_function(results_of_t)
            elif require_all_aggregators:
                raise ValueError(f"No aggregation function for {key=}")
        return merged_results

    @staticmethod
    def flatten_sequence(arrs: list[list]):
        """Flatten a list of lists into a list

        Parameters
        ----------
        arrs : list[list]
            list of lists

        Returns
        -------
        list
            flattened list
        """
        return [item for sublist in arrs for item in sublist]

    @staticmethod
    def ndarray_sum(arrs: list[np.ndarray]):
        """sums an ndarray along ``axis=0``

        Parameters
        ----------
        arrs : list[np.ndarray]
            list of input arrays. Must have the same shape.

        Returns
        -------
        np.ndarray
            sum of input arrays
        """
        return np.array(arrs).sum(axis=0)

    @staticmethod
    def ndarray_mean(arrs: list[np.ndarray]):
        """calculates mean of input ndarrays along ``axis=0``

        Parameters
        ----------
        arrs : list[np.ndarray]
            list of input arrays. Must have the same shape.

        Returns
        -------
        np.ndarray
            mean of input arrays
        """
        return np.array(arrs).mean(axis=0)

    @staticmethod
    def float_mean(floats: list[float]):
        """calculates mean of input float values

        Parameters
        ----------
        floats : list[float]
            list of float values

        Returns
        -------
        float
            mean value
        """
        return np.array(floats).mean()

    @staticmethod
    def ndarray_hstack(arrs: list[np.ndarray]):
        """Performs horizontal stack of input arrays

        Parameters
        ----------
        arrs : list[np.ndarray]
            input numpy arrays

        Returns
        -------
        np.ndarray
            result of stacking
        """
        return np.hstack(arrs)

    @staticmethod
    def ndarray_vstack(arrs: list[np.ndarray]):
        """Performs vertical stack of input arrays

        Parameters
        ----------
        arrs : list[np.ndarray]
            input numpy arrays

        Returns
        -------
        np.ndarray
            result of stacking
        """
        return np.vstack(arrs)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Auxiliary Readers --- :mod:`MDAnalysis.auxiliary`
=================================================

The auxiliary submodule contains code for reading 'auxiliary' data from a 
trajectory and allowing alignment with trajectory timesteps. Additional 
methods in :class:`MDAnalysis.coordinates.base.ProtoReader` allow auxiliary data
to be added and read alongside a trajectory.

Auxiliary data are timeseries accompanying a trajectory not stored in the 
regular trajectory file to be read by the trajectory Reader. They may be stored 
internally (e.g. in an array) or read from a file. In general, auxiliary data is 
assumed to be time ordered and contain no duplicates.

Supported formats
-----------------
Currently supported formats:

   +-------------------------+--------+-----------+---------------------------------------+
   | Reader                  | Format | Extension | Remarks                               |
   |                         |        | (if file) |                                       |
   +=========================+========+===========+=======================================+
   | :class:`.XVGReader`     | XVG    | xvg       | Produced by Gromacs during simulation | 
   |                         |        |           | or analysis.                          |
   |                         |        | (default) | Reads full file on initialisation.    |
   +-------------------------+--------+-----------+---------------------------------------+
   | :class:`.XVGFileReader` | XVG-F  | xvg       | Alternate xvg file reader, reading    |
   |                         |        |           | each step from the file in turn for a |
   |                         |        |           | lower memory footprint.               |
   |                         |        |           | :class:`.XVGReader` is the default    |
   |                         |        |           | reader for .xvg files.                |
   +-------------------------+--------+-----------+---------------------------------------+
   | :class:`.EDRReader`     | EDR    | edr       | Produced by Gromacs during simulation.|
   |                         |        |           | Reads full file on initialisation     |
   +-------------------------+--------+-----------+---------------------------------------+

.. _Auxiliary API:

Auxiliary API
-------------
Auxiliary readers inherit from the base 
:class:`~MDAnalysis.auxiliary.base.AuxReader`. In stand-alone use they 
allow iteration over each step in a set of auxiliary data::

    aux = MDAnalysis.auxiliary.XVG.XVGReader('auxdata.xvg')
    for auxstep in aux:
        print(auxstep)

To iterate over only certain sections of the auxiliary::

    for auxstep in aux[100:200]:  
        # only steps 100-200
        do_something(auxstep)

Or to skip steps::

    for auxstep in aux[100::10]:  
        # every 10th step starting at 100
        do_something(auxstep)

A base :class:`~MDAnalysis.auxiliary.base.AuxFileReader` is also provided,
extending :class:`~MDAnalysis.auxiliary.base.AuxReader` with attributes/methods 
for when auxiliary data is to be read from a file by keeping the file open and
reading steps one-at-time as needed.

A :func:`~MDAnalysis.auxiliary.core.get_auxreader_for` function is available to 
return an appropriate :class:`~MDAnalysis.auxiliary.base.AuxReader` class for
a supplied format or set of auxiliary data, guessing the format from the 
datatype/file extension::

  auxreader = MDAnalysis.auxiliary.core.get_auxreader_for('auxdata.xvg')
  # will return the default XVGReader class
  auxreader = MDAnalysis.auxiliary.core.get_auxreader_for(format='XVG-F')
  # will return the XVGFileReader class

To directly load an instance of the guessed auxiliary reader class given the
supplied auxdata and any additional auxiliary reader options, the 
function :func:`~MDAnalysis.auxiliary.core.auxreader` can be used::

  aux = MDAnalysis.auxiliary.auxreader('auxdata.xvg', dt=2)
  for auxstep in aux:
     do_something(auxstep)


Auxiliaries and trajectories
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
For alignment with trajectory data, auxiliary readers provide methods to 
assign each auxiliary step to the nearest trajectory timestep, read all steps
assigned to a trajectory timestep and calculate 'representative' value(s) of 
the auxiliary data for that timestep. 

'Assignment' of auxiliary steps to trajectory timesteps is determined from the time 
of the auxiliary step, ``dt`` of the trajectory and time at the first frame of the 
trajectory (obtained through a :class:`~MDAnalysis.coordinates.timestep.Timestep` 
instance from the trajectory), as::

    frame = floor((time_at_step - time_at_frame_0 + dt/2)/dt)

If there are no auxiliary steps assigned to a given timestep (or none within 
``cutoff``, if set), the representative value(s) are set to ``np.nan``.

Adding an auxiliary to a trajectory
...................................
Auxiliary data may be added to a trajectory Reader through the 
:meth:`~MDAnalysis.coordinates.base.ProtoReader.add_auxiliary` method. Auxiliary data
may be passed in as a AuxReader instance, or directly as e.g. a filename, in 
which case :func:`~MDAnalysis.auxiliary.core.get_auxreader_for` is used to 
guess an appropriate reader. e.g.::

    u = MDAnalysis.Universe(PDB, XTC)
    u.trajectory.add_auxiliary('pullforce', './pull_force.xvg')

As the trajectory frame is updated, the auxiliary data will be read 
correspondingly, 
updated, and the representative auxiliary value(s) will be available as e.g.
``u.trajectory.ts.aux.pullforce``.


Iterating by an auxiliary
.........................
The trajectory :class:`~MDAnalysis.coordinates.base.ProtoReader` methods 
:meth:`~MDAnalysis.coordinates.base.ProtoReader.next_as_aux` and
:meth:`~MDAnalysis.coordinates.base.ProtoReader.iter_as_aux` allow for movement
through only trajectory timesteps to which one or more steps that fall within 
``cutoff`` from a given auxiliary have been assigned. This may be used to 
avoid representative values set to ``np.nan``, particularly when auxiliary data
is less frequent::

    u.trajectory.add_auxiliary('low_f', 'low_freq_aux_data.xvg')
    for ts in u.trajectory.iter_as_aux('low_f'):
        do_something(ts.aux.low_f) # do something with 'low_f' auxiliary data without
                                   # worrying about having to deal with np.nan

If the auxiliary data are more frequent and the cutoff (if set) is 
sufficiently high, :meth:`~MDAnalysis.coordinates.base.ProtoReader.next_as_aux` 
and :meth:`~MDAnalysis.coordinates.base.ProtoReader.iter_as_aux` behave the same 
as the :class:`~MDAnalysis.coordinates.base.ProtoReader` 
:meth:`~MDAnalysis.coordinates.base.ProtoReader.next` and 
:meth:`~MDAnalysis.coordinates.base.ProtoReader.__iter__`.

In order to access auxiliary values at every individual step, including those 
outside the time range of the trajectory, 
:meth:`~MDAnalysis.coordinates.base.ProtoReader.iter_auxiliary` allows iteration
over the auxiliary independent of the trajectory::

    for auxstep in u.trajectory.iter_auxiliary('pullforce'):
        do_something(auxstep)

To iterate over only a certain section of the auxiliary::

    for auxstep in u.trajectory.iter_auxiliary('pullforce', start=100, step=10):
        # every 10th step from 100
        do_something(auxstep)

The trajectory remains unchanged, and the auxiliary will be returned to the current 
timestep after iteration is complete.

Accessing auxiliary attributes
..............................
To check the values of attributes of an added auxiliary, use 
:meth:`~MDAnalysis.coordinates.base.ProtoReader.get_aux_attribute`, e.g.::

    pullf_dt = u.trajectory.get_aux_attribute('pullforce', 'dt')

If attributes are settable, they can be changed using
:meth:`~MDAnalysis.coordinates.base.ProtoReader.set_aux_attribute`, e.g.::

    u.trajectory.set_aux_attribute('pullforce', 'data_selector', [1])

The auxiliary may be renamed using ``set_aux_attribute`` with 'auxname', or more
directly by using :meth:`~MDAnalysis.coordinates.base.ProtoReader.rename_aux`::

    u.trajectory.rename_aux('pullforce', 'pullf')


Recreating auxiliaries
~~~~~~~~~~~~~~~~~~~~~~~
To recreate an auxiliary, the set of attributes necessary to replicate it can 
first be obtained with :meth:`~MDAnalysis.auxiliary.base.AuxReader.get_description`. 
The returned dictionary can then be passed to 
:func:`~MDAnalysis.auxiliary.core.auxreader` to load a new copy of the 
original auxiliary reader::

    description = aux.get_description()
    del(aux)
    reload_aux = MDAnalysis.auxiliary.auxreader(**description)

The 'description' of any or all the auxiliaries added to a trajectory can be
obtained using :meth:`~MDAnalysis.coordinates.base.ProtoReader.get_aux_descriptions`::

    descriptions = u.trajectory.get_aux_descriptions()

Get descriptions for selected auxiliaries only::

    descriptions = u.trajectory.get_aux_descriptions(['pullf', 'pullx'])

And to reload::

    for descr in descriptions:
        new_u.new_trajectory.add_auxiliary(**descr)


.. _AuxStep API:

AuxStep class
~~~~~~~~~~~~~
An AuxStep instance holds the auxiliary data for the current step. It is
updated whenever a new auxiliary step is read.

AuxStep classes are derived from the base class 
:class:`~MDAnalysis.auxiliary.base.AuxStep`. The appropriate AuxStep class for
a given auxiliary reader is identified in the reader by the `_Auxstep` attribute.


Attributes
..........
The following are inherited from :class:`~MDAnalysis.auxiliary.base.AuxStep`:

  ``step``
      Current auxiliary step (0-based).
  ``_data``
      All recorded data for the current step, as a numpy array.
  ``time``
      Time of current auxiliary step, as a float (in ps). Determined from the 
      ``_data`` if time selection is enabled and a valid ``time_selector`` 
      provided; otherwise calculated using ``dt`` and ``initial_time``.
  ``data``
      Auxiliary values of interest for the current step, as a numpy array. 
      Determined from ``_data`` id data selection is enabled at a valid 
      ``data_selector`` is provided; otherwise set to ``_data``.

The following are stored in AuxStep. The parent 
auxiliary reader has the equivalent attributes, though non-private (no _)
(see :ref:`AuxReader API` below). 

  ``_dt``
      Change in time between auxiliary steps (in ps). If not specified, will
      attempt to determine from auxiliary data; otherwise defaults to 1 ps.
  ``_initial_time``
      Time of first auxiliary step (in ps). If not specified, will attempt to
      determine from auxiliary data; otherwise defaults to 0 ps.
  ``_time_selector``
      Selection key to get time from full set of auxiliary data read with each
      step(``_data``) (ignored if time selection is not enabled by the reader). 
      Type depends on the auxiliary format - e.g. where data is stored in 
      columns, time_selector may be an index of 'time' column. 
      Default value is ``None``, in which case step time is calculated from
      ``dt``, ``initial_time`` and ``step``.
  ``_data_selector``
      Selection key(s) to get data of interest from full set of auxiliary data 
      read with each step (``_data``) (ignored if data selection is not enabled
      by the reader). As for ``time_selector``, type depends on the auxiliary 
      format. If ``None`` (default value), ``_data`` is returned.
  ``_constant_dt``
      Boolean of whether dt is constant throughout auxiliary data. Default is 
      ``True``.

Methods
.......
The following methods are inherited from 
:class:`~MDAnalysis.auxiliary.base.AuxStep`:

  ``__init__(**kwargs)``
    Setup appropriate attributes based on *kwargs*.

To enabled time/data selection in a particular AusStep, the following must be
provided:

  ``_select_time(key)``
    Return, as a float, the time value indicated by *key* from ``_data`` (the full 
    set of data read in from the current step). Raise ``ValueError`` if *key* is 
    not a valid time selector for the auxiliary format.

  ``_select_data(key)``
    Return, as a ndarray, the value(s) indicated by *key* (may be e.g. a list of 
    multiple individual 'keys') from ``_data``. Raise ``ValueError`` if *key* is 
    not a valid data selector for the auxiliary format.

Depending on the format of the auxiliary and hence the format of ``data``, it 
may be necessary to overload the following for a particular AuxStep:

  ``_empty_data()``
    Return a np.array in the same format as ``data`` with all values set to
    ``np.nan``. Used as the representative auxiliary value for a trajectory when 
    no auxiliary steps are assigned to the current frame.


.. _AuxReader API:

AuxReader class
~~~~~~~~~~~~~~~

Registry
........

In order to facilitate guessing of appropriate AuxReaders, all AuxReaders
should set as appropriate `format` attribute. For files, this will be the
expected file extension (in all caps).


Replicating auxiliaries
.......................

The names of the necessary attributes for replicating an auxiliary are stored in 
`required_attrs`, initially set in the base AuxReader. If a particular
AuxReader introduces additional attributes required to reload an auxiliary, 
these should be added.


Attributes
..........

The following attributes are inherited from 
:class:`~MDAnalysis.auxiliary.base.AuxReader`:

  ``auxname``
      Name under which auxiliary data will be stored in trajectory.
  ``represent_ts_as``
      Method to use in calculating representative auxiliary value for a 
      timestep. Default is 'closest'.
  ``cutoff``
      Cutoff (in ps) for ignoring auxiliary steps when calculating
      representative value(s) for a timestep.
  ``auxstep``
      The :class:`~MDAnalysis.auxiliary.base.AuxStep` object to store data for
      the current step. The particular AuxStep used will depend on the auxiliary
      format.
  ``n_steps``
      Total number of auxiliary steps.
  ``frame_data``
      `data` from each auxiliary step assigned to the last-read trajectory
      timestep.
  ``frame_rep``
      Representative value(s) of auxiliary data for last-read trajectory
      timestep.

The following are stored in ``auxstep`` as private attributes (with _)
but may be accessed from the auxiliary reader; see the :ref:`AuxStep API` above.

  ``step``

  ``time``

  ``dt``

  ``initial_time``

  ``time_selector``

  ``data_selector``

  ``constant_dt``

:class:`~MDAnalysis.auxiliary.base.AuxFileReader` provides:

  ``auxfile``
      File object for the auxiliary file.

Each auxiliary reader subclass will additionally set:

   ``_auxdata``
       Value of 'auxdata' used to load the auxiliary - e.g. path to the file 
       containing the auxiliary data. Will be used when getting the 'description'
       of the auxiliary for recreating.

Each auxiliary reader class must also identify an appropriate 
:class:`~MDAnalysis.auxiliary.base.AuxStep` with the `_Auxstep` class attribute
 

Methods
.......
The following methods are inherited from 
:class:`~MDAnalysis.auxiliary.base.AuxReader`:

  ``__init__(**kwargs)``
    Setup appropriate attributes based on *kwargs*.

  ``__len__()``
    Number of steps in auxiliary data.

  ``next()``
    Advance to next step.

  ``__iter__()``
    Allow iteration through each auxiliary step.

  ``__getitem__(step)``
    If *step* is a single index, move to that step and return the :class:`AuxStep`;
    if a list or slice, return an iterator over the specified auxiliary steps.
    See examples in :ref:`Auxiliary API` above.

  ``rewind()``
    Reposition to first step.

  ``update_ts(ts)``
    Read auxiliary steps assigned to trajectory step `ts`, calculate and update 
    the representative value in `ts`. Return `ts`.

  ``read_ts(ts)``
    Read auxiliary steps assigned to trajectory step `ts` and calculate
    representative value.

  ``move_to_ts(ts)``
    Move to the auxiliary step before the trajectory timestep *ts*, such that
    calling ``_read_next_step`` reads the first step assigned to *ts* or, when 
    auxiliary data are less frequent and no steps are assigned to *ts*, the 
    first step following *ts*.

  ``step_to_frame(step, ts, return_time_diff=False)``
    Return the frame number of the trajectory described by `ts` to which 
    the auxiliary step `step` is assigned. Optionally also return the difference
    it time between the step and the returned frame.

  ``step_to_time(step)``
    Return the time of the auxiliary step `step`. 

  ``next_nonempty_frame(ts)``
    Return the frame number of the next trajectory frame (after the current
    auxiliary time) for which a representative auxiliary value can be
    calculated (i.e., there is at least one assigned auxiliary step within
    ``cutoff``).

  ``calc_representative()``
    Return the representative value calculated from ``data`` following
    the method specified by ``represent_ts_as`` and ``cutoff``.

  ``__enter__()``
    Entry method for Context Manager (returns self).

  ``__exit__()``
    Exit method for Context Manager (calls ``close()``).

  ``__del__()``
    Calls ``close()``.

  ``get_description``
    Get the values of the attributes required for replicating an auxiliary (as
    listed in ``required_attrs``) and return as a dictionary.

  ``__eq__``
    Check for equality by checking each of the attributes required for
    replicating an auxiliary (as listed in ``required_attrs``) are equal.

Each AuxReader must subclass :class:`~MDAnalysis.auxiliary.base.AuxReader`
and addionally define:

  ``__init__(auxdata, **kwargs)``
    Additional processing of *kwargs* and any necessary initial processing of 
    *auxdata*. Must :func:`super` through
    :class:`~MDAnalysis.auxiliary.base.AuxReader`.

  ``_read_next_step()``
    Read the data from the next auxiliary step and update ``auxstep`` as
    appropriate. Raise `StopIteration` when attempting to read past last step.

  ``_go_to_step(i)``
    Move to and read step `i` (0-based) from the auxiliary data. Raise 
    ValueError when i is out of range. Update the ``auxstep`` and return.


Depending on the format of the auxiliary data, it may also be necessary to 
define/overwrite the following:

  ``read_all_times()``
    Return the list of times for each step (only required if ``constant_dt``
    may be false).

  ``count_n_steps()``
    Return the total number of steps (only required if `_n_steps` not otherwise
    set during `__init__`).

  ``_restart()``
    Reposition before the first step

  ``close()``

For convenience, when reading auxiliary data from an open file, one step at a
time, :class:`~MDAnalysis.auxiliary.base.AuxFileReader`
extends :class:`~MDAnalysis.auxiliary.base.AuxReader` by providing the
following (these may be overwritten by subclasses as appropriate):

  ``__init__(auxfile, **kwargs)``
    Open *auxfile* and any additional processing of *kwargs*.

  ``_restart()``,
    Seek to the start of ``auxfile``

  ``close()``
    Close ``auxfile``

  ``_go_to_step(i)``
    Iterate through all steps until step `i` is reached.

  ``_reopen()``
    Close ``auxfile`` and reopen.


"""

# registry of auxiliary readers
_AUXREADERS = {}

from . import base
from . import XVG
from . import EDR
from .core import auxreader



"""Analysis backends --- :mod:`MDAnalysis.analysis.backends`
============================================================

.. versionadded:: 2.8.0


The :mod:`backends` module provides :class:`BackendBase` base class to
implement custom execution backends for
:meth:`MDAnalysis.analysis.base.AnalysisBase.run` and its
subclasses.

.. SeeAlso:: :ref:`parallel-analysis`

.. _backends:

Backends
--------

Three built-in backend classes are provided:

* *serial*: :class:`BackendSerial`, that is equivalent to using no
  parallelization and is the default

* *multiprocessing*: :class:`BackendMultiprocessing` that supports
  parallelization via standard Python :mod:`multiprocessing` module
  and uses default :mod:`pickle` serialization

* *dask*: :class:`BackendDask`, that uses the same process-based
  parallelization as :class:`BackendMultiprocessing`, but different
  serialization algorithm via `dask <https://dask.org/>`_ (see `dask
  serialization algorithms
  <https://distributed.dask.org/en/latest/serialization.html>`_ for details)

Classes
-------

"""

import warnings
from typing import Callable
from MDAnalysis.lib.util import is_installed


class BackendBase:
    """Base class for backend implementation.

    Initializes an instance and performs checks for its validity, such as
    ``n_workers`` and possibly other ones.

    Parameters
    ----------
    n_workers : int
        number of workers (usually, processes) over which the work is split

    Examples
    --------
    .. code-block:: python

        from MDAnalysis.analysis.backends import BackendBase

        class ThreadsBackend(BackendBase):
            def apply(self, func, computations):
                from multiprocessing.dummy import Pool

                with Pool(processes=self.n_workers) as pool:
                    results = pool.map(func, computations)
                return results

        import MDAnalysis as mda
        from MDAnalysis.tests.datafiles import PSF, DCD
        from MDAnalysis.analysis.rms import RMSD

        u = mda.Universe(PSF, DCD)
        ref = mda.Universe(PSF, DCD)

        R = RMSD(u, ref)

        n_workers = 2
        backend = ThreadsBackend(n_workers=n_workers)
        R.run(backend=backend, unsupported_backend=True)

    .. warning::
        Using `ThreadsBackend` above will lead to erroneous results, since it
        is an educational example. Do not use it for real analysis.


    .. versionadded:: 2.8.0

    """

    def __init__(self, n_workers: int):
        self.n_workers = n_workers
        self._validate()

    def _get_checks(self):
        """Get dictionary with ``condition: error_message`` pairs that ensure the
        validity of the backend instance

        Returns
        -------
        dict
            dictionary with ``condition: error_message`` pairs that will get
            checked during ``_validate()`` run
        """
        return {
            isinstance(self.n_workers, int)
            and self.n_workers
            > 0: f"n_workers should be positive integer, got {self.n_workers=}",
        }

    def _get_warnings(self):
        """Get dictionary with ``condition: warning_message`` pairs that ensure
        the good usage of the backend instance

        Returns
        -------
        dict
            dictionary with ``condition: warning_message`` pairs that will get
            checked during ``_validate()`` run
        """
        return dict()

    def _validate(self):
        """Check correctness (e.g. ``dask`` is installed if using ``backend='dask'``)
        and good usage (e.g. ``n_workers=1`` if backend is serial) of the backend

        Raises
        ------
        ValueError
            if one of the conditions in :meth:`_get_checks` is ``True``
        """
        for check, msg in self._get_checks().items():
            if not check:
                raise ValueError(msg)
        for check, msg in self._get_warnings().items():
            if not check:
                warnings.warn(msg)

    def apply(self, func: Callable, computations: list) -> list:
        """map function `func` to all tasks in the `computations` list

        Main method that will get called when using an instance of
        ``BackendBase``.  It is equivalent to running ``[func(item) for item in
        computations]`` while using the parallel backend capabilities.

        Parameters
        ----------
        func : Callable
            function to be called on each of the tasks in computations list
        computations : list
            computation tasks to apply function to

        Returns
        -------
        list
            list of results of the function

        """
        raise NotImplementedError


class BackendSerial(BackendBase):
    """A built-in backend that does serial execution of the function, without any
    parallelization.

    Parameters
    ----------
    n_workers : int
        Is ignored in this class, and if ``n_workers`` > 1, a warning will be
        given.


    .. versionadded:: 2.8.0
    """

    def _get_warnings(self):
        """Get dictionary with ``condition: warning_message`` pairs that ensure
        the good usage of the backend instance. Here, it checks if the number
        of workers is not 1, otherwise gives warning.

        Returns
        -------
        dict
            dictionary with ``condition: warning_message`` pairs that will get
            checked during ``_validate()`` run
        """
        return {
            self.n_workers
            == 1: "n_workers is ignored when executing with backend='serial'"
        }

    def apply(self, func: Callable, computations: list) -> list:
        """
        Serially applies `func` to each task object in ``computations``.

        Parameters
        ----------
        func : Callable
            function to be called on each of the tasks in computations list
        computations : list
            computation tasks to apply function to

        Returns
        -------
        list
            list of results of the function
        """
        return [func(task) for task in computations]


class BackendMultiprocessing(BackendBase):
    """A built-in backend that executes a given function using the
    :meth:`multiprocessing.Pool.map <multiprocessing.pool.Pool.map>` method.

    Parameters
    ----------
    n_workers : int
        number of processes in :class:`multiprocessing.Pool
        <multiprocessing.pool.Pool>` to distribute the workload
        between. Must be a positive integer.

    Examples
    --------

    .. code-block:: python

        from MDAnalysis.analysis.backends import BackendMultiprocessing
        import multiprocessing as mp

        backend_obj = BackendMultiprocessing(n_workers=mp.cpu_count())


    .. versionadded:: 2.8.0

    """

    def apply(self, func: Callable, computations: list) -> list:
        """Applies `func` to each object in ``computations`` using `multiprocessing`'s `Pool.map`.

        Parameters
        ----------
        func : Callable
            function to be called on each of the tasks in computations list
        computations : list
            computation tasks to apply function to

        Returns
        -------
        list
            list of results of the function
        """
        from multiprocessing import Pool

        with Pool(processes=self.n_workers) as pool:
            results = pool.map(func, computations)
        return results


class BackendDask(BackendBase):
    """A built-in backend that executes a given function with *dask*.

    Execution is performed with the :func:`dask.compute` function of
    :class:`dask.delayed.Delayed` object (created with
    :func:`dask.delayed.delayed`) with ``scheduler='processes'`` and
    ``chunksize=1`` (this ensures uniform distribution of tasks among
    processes). Requires the `dask package <https://docs.dask.org/en/stable/>`_
    to be `installed <https://docs.dask.org/en/stable/install.html>`_.

    Parameters
    ----------
    n_workers : int
        number of processes in to distribute the workload
        between. Must be a positive integer. Workers are actually
        :class:`multiprocessing.pool.Pool` processes, but they use a different and
        more flexible `serialization protocol
        <https://docs.dask.org/en/stable/phases-of-computation.html#graph-serialization>`_.

    Examples
    --------

    .. code-block:: python

        from MDAnalysis.analysis.backends import BackendDask
        import multiprocessing as mp

        backend_obj = BackendDask(n_workers=mp.cpu_count())


    .. versionadded:: 2.8.0

    """

    def apply(self, func: Callable, computations: list) -> list:
        """Applies `func` to each object in ``computations``.

        Parameters
        ----------
        func : Callable
            function to be called on each of the tasks in computations list
        computations : list
            computation tasks to apply function to

        Returns
        -------
        list
            list of results of the function
        """
        from dask.delayed import delayed
        import dask

        computations = [delayed(func)(task) for task in computations]
        results = dask.compute(
            computations,
            scheduler="processes",
            chunksize=1,
            num_workers=self.n_workers,
        )[0]
        return results

    def _get_checks(self):
        """Get dictionary with ``condition: error_message`` pairs that ensure the
        validity of the backend instance. Here checks if ``dask`` module is
        installed in the environment.

        Returns
        -------
        dict
            dictionary with ``condition: error_message`` pairs that will get
            checked during ``_validate()`` run
        """
        base_checks = super()._get_checks()
        checks = {
            is_installed("dask"): (
                "module 'dask' is missing. Please install 'dask': "
                "https://docs.dask.org/en/stable/install.html"
            )
        }
        return base_checks | checks



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
MMTF Topology Parser
====================

Reads topology data from the `Macromolecular Transmission Format
(MMTF) format`_.  This should generally be a quicker alternative to PDB.

Makes individual models within the MMTF file available via the `models`
attribute on Universe.

.. versionadded:: 0.16.0
.. versionchanged:: 0.20.0
   Can now read files with optional fields missing/empty

.. versionchanged:: 2.0.0
   Aliased ``bfactors`` topologyattribute to ``tempfactors``.
   ``tempfactors`` is deprecated and will be removed in 3.0 (Issue #1901)
.. versionchanged:: 2.8.0
    Removed mass guessing (attributes guessing takes place now
    through universe.guess_TopologyAttrs() API).

Reads the following topology attributes:

Atoms:
 - altLoc
 - atom ID
 - tempfactor
 - bonds
 - charge
 - name
 - occupancy
 - type

Residues:
 - icode
 - resname
 - resid
 - resnum

Segments:
 - segid
 - model

 .. note::

    By default, masses will be guessed on Universe creation.
    This may change in release 3.0.
    See :ref:`Guessers` for more information.

Classes
-------

.. autoclass:: MMTFParser
   :members:

.. _Macromolecular Transmission Format (MMTF) format: https://www.rcsb.org/news/feature/65a1af31c76ca3abcc925d0c
"""
from collections import defaultdict
import mmtf
import numpy as np


from . import base
from ..core.topology import Topology
from ..core.topologyattrs import (
    AltLocs,
    Atomids,
    Atomnames,
    Atomtypes,
    Tempfactors,
    Bonds,
    Charges,
    ICodes,
    Occupancies,
    Resids,
    Resnames,
    Resnums,
    Segids,
    SegmentAttr,  # for model
)
from ..core.selection import RangeSelection
from ..due import due, Doi


def _parse_mmtf(fn):
    if fn.endswith("gz"):
        return mmtf.parse_gzip(fn)
    else:
        return mmtf.parse(fn)


class Models(SegmentAttr):
    attrname = "models"
    singular = "model"
    transplants = defaultdict(list)

    def models(self):
        """Models in this Universe.

        The MMTF format can define various models for a given structure. The
        topology (eg residue identity) can change between different models,
        resulting in a different number of atoms in each model.

        Returns
        -------
        A list of AtomGroups, each representing a single model.
        """
        model_ids = np.unique(self.segments.models)

        return [self.select_atoms("model {}".format(i)) for i in model_ids]

    transplants["Universe"].append(
        ("models", property(models, None, None, models.__doc__))
    )


class ModelSelection(RangeSelection):
    token = "model"
    field = "models"

    def apply(self, group):
        mask = np.zeros(len(group), dtype=bool)
        vals = group.models

        for upper, lower in zip(self.uppers, self.lowers):
            if upper is not None:
                thismask = vals >= lower
                thismask &= vals <= upper
            else:
                thismask = vals == lower

            mask |= thismask
        return group[mask].unique


class MMTFParser(base.TopologyReaderBase):
    format = "MMTF"

    @staticmethod
    def _format_hint(thing):
        """Can parser read *thing*?

        .. versionadded:: 1.0.0
        """
        return isinstance(thing, mmtf.MMTFDecoder)

    @due.dcite(
        Doi("10.1371/journal.pcbi.1005575"),
        description="MMTF Parser",
        path="MDAnalysis.topology.MMTFParser",
    )
    def parse(self, **kwargs):
        if isinstance(self.filename, mmtf.MMTFDecoder):
            mtop = self.filename
        else:
            mtop = _parse_mmtf(self.filename)

        def iter_atoms(field):
            # iterate through atoms in groups
            for i in mtop.group_type_list:
                g = mtop.group_list[i]
                for val in g[field]:
                    yield val

        natoms = mtop.num_atoms
        nresidues = mtop.num_groups
        nsegments = mtop.num_chains
        attrs = []

        # required
        charges = Charges(list(iter_atoms("formalChargeList")))
        names = Atomnames(list(iter_atoms("atomNameList")))
        types = Atomtypes(list(iter_atoms("elementList")))
        attrs.extend([charges, names, types])

        # optional are empty list if empty, sometimes arrays
        if len(mtop.atom_id_list):
            attrs.append(Atomids(mtop.atom_id_list))
        else:
            # must have this attribute for MDA
            attrs.append(Atomids(np.arange(natoms), guessed=True))
        if mtop.alt_loc_list:
            attrs.append(
                AltLocs(
                    [
                        val.replace("\x00", "").strip()
                        for val in mtop.alt_loc_list
                    ]
                )
            )
        else:
            attrs.append(AltLocs([""] * natoms))
        if len(mtop.b_factor_list):
            attrs.append(Tempfactors(mtop.b_factor_list))
        else:
            attrs.append(Tempfactors([0] * natoms))
        if len(mtop.occupancy_list):
            attrs.append(Occupancies(mtop.occupancy_list))
        else:
            attrs.append(Occupancies([1] * natoms))

        # Residue things
        # required
        resids = Resids(mtop.group_id_list)
        resnums = Resnums(resids.values.copy())
        resnames = Resnames(
            [mtop.group_list[i]["groupName"] for i in mtop.group_type_list]
        )
        attrs.extend([resids, resnums, resnames])
        # optional
        # mmtf empty icode is '\x00' rather than ''
        if mtop.ins_code_list:
            attrs.append(
                ICodes(
                    [
                        val.replace("\x00", "").strip()
                        for val in mtop.ins_code_list
                    ]
                )
            )
        else:
            attrs.append(ICodes([""] * nresidues))

        # Segment things
        # optional
        if mtop.chain_name_list:
            attrs.append(Segids(mtop.chain_name_list))
        else:
            # required for MDAnalysis
            attrs.append(Segids(["SYSTEM"] * nsegments, guessed=True))

        mods = np.repeat(np.arange(mtop.num_models), mtop.chains_per_model)
        attrs.append(Models(mods))
        # attrs.append(chainids)

        # number of atoms in a given group id
        groupID_2_natoms = {
            i: len(g["atomNameList"]) for i, g in enumerate(mtop.group_list)
        }
        # mapping of atoms to residues
        resindex = np.repeat(
            np.arange(nresidues),
            [groupID_2_natoms[i] for i in mtop.group_type_list],
        )

        # mapping of residues to segments
        segindex = np.repeat(np.arange(nsegments), mtop.groups_per_chain)

        # Bonds
        # bonds are listed as indices within a group,
        # offset pulls out 'global' index
        offset = 0
        bonds = []
        for gtype in mtop.group_type_list:
            g = mtop.group_list[gtype]
            bondlist = g["bondAtomList"]

            for x, y in zip(bondlist[1::2], bondlist[::2]):
                if x > y:
                    x, y = y, x  # always have x < y
                bonds.append((x + offset, y + offset))

            offset += groupID_2_natoms[gtype]
        # add inter group bonds
        if not mtop.bond_atom_list is None:  # optional field
            for x, y in zip(
                mtop.bond_atom_list[1::2], mtop.bond_atom_list[::2]
            ):
                if x > y:
                    x, y = y, x
                bonds.append((x, y))
        attrs.append(Bonds(bonds))

        top = Topology(
            natoms,
            nresidues,
            nsegments,
            atom_resindex=resindex,
            residue_segindex=segindex,
            attrs=attrs,
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numbers
from multiprocessing.sharedctypes import SynchronizedArray
from multiprocessing import Process, Manager
from joblib import cpu_count
import numpy as np
import sys

import MDAnalysis as mda
from ...coordinates.memory import MemoryReader


class TriangularMatrix(object):
    """Triangular matrix class. This class is designed to provide a
    memory-efficient representation of a triangular matrix that still behaves
    as a square symmetric one. The class wraps a numpy.array object,
    in which data are memorized in row-major order. It also has few additional
    facilities to conveniently load/write a matrix from/to file. It can be
    accessed using the [] and () operators, similarly to a normal numpy array.

    """

    def __init__(self, size, metadata=None, loadfile=None):
        """Class constructor.

        Parameters
        ----------

        size : int / array_like
            Size of the matrix (number of rows or columns). If an
            array is provided instead, the size of the triangular matrix
            will be calculated and the array copied as the matrix
            elements. Otherwise, the matrix is just initialized to zero.
        metadata : dict or None
            Metadata dictionary. Used to generate the metadata attribute.
        loadfile : str or None
            Load the matrix from this file. All the attributes and data will
            be determined by the matrix file itself (i.e. metadata will be
            ignored); size has to be provided though.

        """
        if isinstance(metadata, dict):
            self.metadata = np.array(metadata.items(), dtype=object)
        else:
            self.metadata = metadata

        self.size = size
        if loadfile:
            self.loadz(loadfile)
        elif isinstance(size, numbers.Integral):
            self.size = size
            self._elements = np.zeros((size + 1) * size // 2, dtype=np.float64)
        elif isinstance(size, SynchronizedArray):
            self._elements = np.array(size.get_obj(), dtype=np.float64)
            self.size = int((np.sqrt(1 + 8 * len(size)) - 1) / 2)
        elif isinstance(size, np.ndarray):
            self._elements = size
            self.size = int((np.sqrt(1 + 8 * len(size)) - 1) / 2)
        else:
            raise TypeError

    def __getitem__(self, args):
        x, y = args
        if x < y:
            x, y = y, x
        return self._elements[x * (x + 1) // 2 + y]

    def __setitem__(self, args, val):
        x, y = args
        if x < y:
            x, y = y, x
        self._elements[x * (x + 1) // 2 + y] = val

    def as_array(self):
        """Return standard numpy array equivalent"""
        # pylint: disable=unsubscriptable-object
        a = np.zeros((self.size, self.size))
        a[np.tril_indices(self.size)] = self._elements
        a[np.triu_indices(self.size)] = a.T[np.triu_indices(self.size)]
        return a

    def savez(self, fname):
        """Save matrix in the npz compressed numpy format. Save metadata and
        data as well.

        Parameters
        ----------

        fname : str
            Name of the file to be saved.
        """
        np.savez(fname, elements=self._elements, metadata=self.metadata)

    def loadz(self, fname):
        """Load matrix from the npz compressed numpy format.

        Parameters
        ----------

        fname : str
            Name of the file to be loaded.
        """
        loaded = np.load(fname, allow_pickle=True)

        if loaded["metadata"].shape != ():
            if loaded["metadata"]["number of frames"] != self.size:
                raise TypeError
            self.metadata = loaded["metadata"]
        else:
            if self.size * (self.size - 1) / 2 + self.size != len(
                loaded["elements"]
            ):
                raise TypeError
        self._elements = loaded["elements"]

    def __add__(self, scalar):
        """Add scalar to matrix elements.

        Parameters
        ----------

        scalar : float
            Scalar to be added.
        """
        newMatrix = self.__class__(self.size)
        newMatrix._elements = self._elements + scalar
        return newMatrix

    def __iadd__(self, scalar):
        """Add scalar to matrix elements.

        Parameters
        ----------

        scalar : float
            Scalar to be added.
        """
        self._elements += scalar
        return self

    def __mul__(self, scalar):
        """Multiply with scalar.

        Parameters
        ----------

        scalar : float
            Scalar to multiply with.
        """
        newMatrix = self.__class__(self.size)
        newMatrix._elements = self._elements * scalar
        return newMatrix

    def __imul__(self, scalar):
        """Multiply with scalar.

        Parameters
        ----------

        scalar : float
            Scalar to multiply with.
        """
        self._elements *= scalar
        return self

    __rmul__ = __mul__

    def __str__(self):
        return str(self.as_array())


class ParallelCalculation(object):
    r"""
    Generic parallel calculation class. Can use arbitrary functions,
    arguments to functions and kwargs to functions.

    Attributes
    ----------
    n_jobs : int
        Number of cores to be used for parallel calculation. If -1 use all
        available cores.
    function : callable object
        Function to be run in parallel.
    args : list of tuples
        Each tuple contains the arguments that will be passed to
        function(). This means that a call to function() is performed for
        each tuple. function is called as function(\*args, \*\*kwargs). Runs
        are distributed on the requested numbers of cores.
    kwargs : list of dicts
        Each tuple contains the named arguments that will be passed to
        function, similarly as described for the args attribute.
    nruns : int
        Number of runs to be performed. Must be equal to len(args) and
        len(kwargs).
    """

    def __init__(self, n_jobs, function, args=None, kwargs=None):
        """
        Parameters
        ----------
        n_jobs : int
            Number of cores to be used for parallel calculation. If -1 use all
            available cores.
        function : object that supports __call__, as functions
            function to be run in parallel.
        args : list of tuples
            Arguments for function; see the ParallelCalculation class
            description.
        kwargs : list of dicts or None
            kwargs for function; see the ParallelCalculation
            class description.
        """

        # args[i] should be a list of args, one for each run
        self.n_jobs = n_jobs
        if self.n_jobs == -1:
            self.n_jobs = cpu_count()

        self.functions = function
        if not hasattr(self.functions, "__iter__"):
            self.functions = [self.functions] * len(args)
        if len(self.functions) != len(args):
            self.functions = self.functions[:] * (
                len(args) // len(self.functions)
            )

        # Arguments should be present
        if args is None:
            args = []
        self.args = args

        # If kwargs are not present, use empty dicts
        if kwargs:
            self.kwargs = kwargs
        else:
            self.kwargs = [{} for i in self.args]

        self.nruns = len(args)

    def worker(self, q, results):
        """
        Generic worker. Will run function with the prescribed args and kwargs.

        Parameters
        ----------

        q : multiprocessing.Manager.Queue object
                work queue, from which the worker fetches arguments and
                messages

        results : multiprocessing.Manager.Queue object
                results queue, where results are put after each calculation is
                finished

        """
        while True:
            i = q.get()
            if i == "STOP":
                return

            results.put(
                (i, self.functions[i](*self.args[i], **self.kwargs[i]))
            )

    def run(self):
        r"""
        Run parallel calculation.

        Returns
        -------

        results : tuple of ordered tuples (int, object)
                int is the number of the calculation corresponding to a
                certain argument in the args list, and object is the result of
                corresponding calculation. For instance, in (3, output), output
                is the return of function(\*args[3], \*\*kwargs[3]).
        """
        results_list = []
        if self.n_jobs == 1:
            for i in range(self.nruns):
                results_list.append(
                    (i, self.functions[i](*self.args[i], **self.kwargs[i]))
                )
        else:
            manager = Manager()
            q = manager.Queue()
            results = manager.Queue()

            workers = [
                Process(target=self.worker, args=(q, results))
                for i in range(self.n_jobs)
            ]

            for i in range(self.nruns):
                q.put(i)
            for w in workers:
                q.put("STOP")

            for w in workers:
                w.start()

            for w in workers:
                w.join()

            results.put("STOP")
            for i in iter(results.get, "STOP"):
                results_list.append(i)

        return tuple(sorted(results_list, key=lambda x: x[0]))


def trm_indices(a, b):
    """
    Generate (i,j) indeces of a triangular matrix, between elements a and b.
    The matrix size is automatically determined from the number of elements.
    For instance: trm_indices((0,0),(2,1)) yields (0,0) (1,0) (1,1) (2,0)
    (2,1).

    Parameters
    ----------

    a : (int i, int j) tuple
        starting matrix element.

    b : (int i, int j) tuple
        final matrix element.
    """
    i, j = a
    while i < b[0]:
        if i == j:
            yield (i, j)
            j = 0
            i += 1
        else:
            yield (i, j)
            j += 1
    while j <= b[1]:
        yield (i, j)
        j += 1


def trm_indices_nodiag(n):
    """generate (i,j) indeces of a triangular matrix of n rows (or columns),
    without diagonal (e.g. no elements (0,0),(1,1),...,(n,n))

    Parameters
    ----------

    n : int
        Matrix size
    """

    for i in range(1, n):
        for j in range(i):
            yield (i, j)


def trm_indices_diag(n):
    """generate (i,j) indeces of a triangular matrix of n rows (or columns),
    with diagonal

    Parameters
    ----------

    n : int
        Matrix size
    """

    for i in range(0, n):
        for j in range(i + 1):
            yield (i, j)


def merge_universes(universes):
    """
    Merge list of universes into one

    Parameters
    ----------
    universes : list of Universe objects


    Returns
    ----------
    Universe object
    """

    for universe in universes:
        universe.transfer_to_memory()

    return mda.Universe(
        universes[0].filename,
        np.concatenate(
            tuple([e.trajectory.timeseries(order="fac") for e in universes]),
            axis=0,
        ),
        format=MemoryReader,
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""Hydrogen bond autocorrelation --- :mod:`MDAnalysis.analysis.hbonds.hbond_autocorrel` (deprecated)
====================================================================================================

:Author: Richard J. Gowers
:Year: 2014
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 0.9.0

.. deprecated:: 2.0.0

   This module was moved to
   :mod:`MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel` and access to this
   module through :mod:`MDAnalysis.analysis.hbonds.hbond_autocorrel` will be
   removed in 3.0.0.


See Also
--------
:mod:`MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel`

"""
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("always", DeprecationWarning)
    wmsg = (
        "This module was moved to "
        "MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel; "
        "hbonds.hbond_autocorrel will be removed in 3.0.0."
    )
    warnings.warn(wmsg, category=DeprecationWarning)

from MDAnalysis.lib.util import deprecate

from ..hydrogenbonds import hbond_autocorrel


find_hydrogen_donors = deprecate(
    hbond_autocorrel.find_hydrogen_donors,
    release="2.0.0",
    remove="3.0.0",
    message="The function was moved to "
    "MDAnalysis.analysis.hbonds.hbond_autocorrel.",
)

HydrogenBondAutoCorrel = deprecate(
    hbond_autocorrel.HydrogenBondAutoCorrel,
    release="2.0.0",
    remove="3.0.0",
    message="The class was moved to "
    "MDAnalysis.analysis.hbonds.hbond_autocorrel.",
)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import MDAnalysis as mda
from MDAnalysis.auxiliary.base import AuxReader

import pytest


def test_get_auxreader_for_none():
    with pytest.raises(
        ValueError, match="Must provide either auxdata or format"
    ):
        mda.auxiliary.core.get_auxreader_for()


def test_get_auxreader_for_wrong_auxdata():
    with pytest.raises(
        ValueError, match="Unknown auxiliary data format for auxdata:"
    ):
        mda.auxiliary.core.get_auxreader_for(auxdata="test.none")


def test_get_auxreader_for_wrong_format():
    with pytest.raises(ValueError, match="Unknown auxiliary data format"):
        mda.auxiliary.core.get_auxreader_for(format="none")


def test_notimplemented_read_next_timestep():
    with pytest.raises(
        NotImplementedError, match="BUG: Override " "_read_next_step()"
    ):
        reader = mda.auxiliary.base.AuxReader()


class No_Memory_Usage(AuxReader):
    def _read_next_step(self):
        pass


def test_notimplemented_memory_usage():
    with pytest.raises(
        NotImplementedError, match="BUG: Override " "_memory_usage()"
    ):
        reader = No_Memory_Usage()
        reader._memory_usage()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
PQR topology parser
===================

Read atoms with charges from a PQR_ file (as written by PDB2PQR_). No
connectivity is deduced.

Note
----
The file format is described in :mod:`MDAnalysis.coordinates.PQR`.


Classes
-------

.. autoclass:: PQRParser
   :members:
   :inherited-members:


.. _PQR:     https://apbs-pdb2pqr.readthedocs.io/en/latest/formats/pqr.html
.. _APBS:    https://apbs.readthedocs.io/en/latest/
.. _PDB2PQR: https://apbs-pdb2pqr.readthedocs.io/en/latest/pdb2pqr/index.html
.. _PDB:     http://www.wwpdb.org/documentation/file-format

"""
import numpy as np

from ..lib.util import openany
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Atomtypes,
    Charges,
    ICodes,
    Radii,
    RecordTypes,
    Resids,
    Resnums,
    Resnames,
    Segids,
)
from ..core.topology import Topology
from .base import TopologyReaderBase, squash_by, change_squash


class PQRParser(TopologyReaderBase):
    """Parse atom information from PQR file *filename*.

    Creates a MDAnalysis Topology with the following attributes
     - Atomids
     - Atomnames
     - Charges
     - Radii
     - RecordTypes (ATOM/HETATM)
     - Resids
     - Resnames
     - Segids

     .. note::

        Atomtypes will be read from the input file if they are present
        (e.g. GROMACS PQR files). Otherwise, they will be guessed on Universe
        creation. By default, masses will also be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.


    .. versionchanged:: 0.9.0
       Read chainID from a PQR file and use it as segid (before we always used
       'SYSTEM' as the new segid).
    .. versionchanged:: 0.16.1
       Now reads insertion codes and splits into new residues around these
    .. versionchanged:: 0.18.0
       Added parsing of Record types
       Can now read PQR files from Gromacs, these provide atom type as last column
       but don't have segids
    .. versionchanged:: 2.8.0
        Removed type and mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).

    """

    format = "PQR"

    @staticmethod
    def guess_flavour(line):
        """Guess which variant of PQR format this line is

        Parameters
        ----------
        line : str
          entire line of PQR file starting with ATOM/HETATM

        Returns
        -------
        flavour : str
          ORIGINAL / GROMACS / NO_CHAINID

        .. versionadded:: 0.18.0
        """
        fields = line.split()
        if len(fields) == 11:
            try:
                float(fields[-1])
            except ValueError:
                flavour = "GROMACS"
            else:
                flavour = "ORIGINAL"
        else:
            flavour = "NO_CHAINID"
        return flavour

    def parse(self, **kwargs):
        """Parse atom information from PQR file *filename*.

        Returns
        -------
        A MDAnalysis Topology object
        """
        record_types = []
        serials = []
        names = []
        resnames = []
        chainIDs = []
        resids = []
        icodes = []
        charges = []
        radii = []
        elements = []

        flavour = None

        with openany(self.filename) as f:
            for line in f:
                if not line.startswith(("ATOM", "HETATM")):
                    continue
                fields = line.split()

                if flavour is None:
                    flavour = self.guess_flavour(line)
                if flavour == "ORIGINAL":
                    (
                        recordName,
                        serial,
                        name,
                        resName,
                        chainID,
                        resSeq,
                        x,
                        y,
                        z,
                        charge,
                        radius,
                    ) = fields
                elif flavour == "GROMACS":
                    (
                        recordName,
                        serial,
                        name,
                        resName,
                        resSeq,
                        x,
                        y,
                        z,
                        charge,
                        radius,
                        element,
                    ) = fields
                    chainID = "SYSTEM"
                    elements.append(element)
                elif flavour == "NO_CHAINID":
                    # files without the chainID
                    (
                        recordName,
                        serial,
                        name,
                        resName,
                        resSeq,
                        x,
                        y,
                        z,
                        charge,
                        radius,
                    ) = fields
                    chainID = "SYSTEM"

                try:
                    resid = int(resSeq)
                except ValueError:
                    # has icode present
                    resid = int(resSeq[:-1])
                    icode = resSeq[-1]
                else:
                    icode = ""

                record_types.append(recordName)
                serials.append(serial)
                names.append(name)
                resnames.append(resName)
                resids.append(resid)
                icodes.append(icode)
                charges.append(charge)
                radii.append(radius)
                chainIDs.append(chainID)

        n_atoms = len(serials)

        attrs = []
        if elements:
            atomtypes = elements
            attrs.append(Atomtypes(atomtypes, False))

        attrs.append(Atomids(np.array(serials, dtype=np.int32)))
        attrs.append(Atomnames(np.array(names, dtype=object)))
        attrs.append(Charges(np.array(charges, dtype=np.float32)))
        attrs.append(RecordTypes(np.array(record_types, dtype=object)))
        attrs.append(Radii(np.array(radii, dtype=np.float32)))

        resids = np.array(resids, dtype=np.int32)
        icodes = np.array(icodes, dtype=object)
        resnames = np.array(resnames, dtype=object)
        chainIDs = np.array(chainIDs, dtype=object)

        residx, (resids, resnames, icodes, chainIDs) = change_squash(
            (resids, resnames, icodes, chainIDs),
            (resids, resnames, icodes, chainIDs),
        )

        n_residues = len(resids)
        attrs.append(Resids(resids))
        attrs.append(Resnums(resids.copy()))
        attrs.append(Resnames(resnames))
        attrs.append(ICodes(icodes))

        segidx, chainIDs = squash_by(chainIDs)[:2]

        n_segments = len(chainIDs)
        attrs.append(Segids(chainIDs))

        top = Topology(
            n_atoms,
            n_residues,
            n_segments,
            attrs=attrs,
            atom_resindex=residx,
            residue_segindex=segidx,
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from contextlib import contextmanager

import MDAnalysis as mda
import MDAnalysis.analysis.align as align
from MDAnalysis.analysis.align import HAS_BIOPYTHON
import MDAnalysis.analysis.rms as rms
import os
import numpy as np
import pytest
from MDAnalysis import SelectionError, SelectionWarning
from MDAnalysisTests import executable_not_found
from MDAnalysisTests.datafiles import (
    PSF,
    DCD,
    CRD,
    FASTA,
    ALIGN_BOUND,
    ALIGN_UNBOUND,
    PDB_helix,
)
from numpy.testing import (
    assert_equal,
    assert_array_equal,
    assert_allclose,
)


# Function for Parametrizing conditional raising
@contextmanager
def does_not_raise():
    yield


class TestRotationMatrix:
    a = np.array([[0.1, 0.2, 0.3], [1.1, 1.1, 1.1]])
    b = np.array([[0.1, 0.1, 0.1], [1.1, 1.1, 1.1]])
    w = np.array([1.3, 2.3])

    @pytest.mark.parametrize(
        "a, b, weights, expected",
        (
            (a, b, None, 0.15785647734415692),
            (a, b, w, 0.13424643502242328),
        ),
    )
    def test_rotation_matrix_input(self, a, b, weights, expected):
        rot, rmsd = align.rotation_matrix(a, b, weights)
        assert_equal(rot, np.eye(3))
        assert rmsd == pytest.approx(expected)

    def test_list_args(self):
        a = [[0.1, 0.2, 0.3], [1.1, 1.1, 1.1]]
        b = [[0.1, 0.1, 0.1], [1.1, 1.1, 1.1]]
        w = [1.3, 2.3]
        rot, rmsd = align.rotation_matrix(a, b, w)
        assert_equal(rot, np.eye(3))
        assert rmsd == pytest.approx(0.13424643502242328)

    def test_exception(self):
        a = [[0.1, 0.2, 0.3], [1.1, 1.1, 1.1], [2, 2, 2]]
        b = [[0.1, 0.1, 0.1], [1.1, 1.1, 1.1]]
        with pytest.raises(ValueError):
            align.rotation_matrix(a, b)


class TestGetMatchingAtoms(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PSF, DCD)

    @staticmethod
    @pytest.fixture()
    def reference():
        return mda.Universe(PSF, DCD)

    @staticmethod
    @pytest.fixture()
    def reference_small(reference):
        return mda.Merge(
            reference.select_atoms("not name H* and not atom 4AKE 1 CA")
        )

    @pytest.mark.parametrize("strict", (True, False))
    def test_match(
        self, universe, reference, strict, selection="protein and backbone"
    ):
        ref = reference.select_atoms(selection)
        mobile = universe.select_atoms(selection)
        groups = align.get_matching_atoms(ref, mobile, strict=strict)
        assert_equal(groups[0].names, groups[1].names)

    @pytest.mark.parametrize("strict", (True, False))
    def test_nomatch_atoms_raise(
        self, universe, reference, strict, selection="protein and backbone"
    ):
        # one atom less but same residues; with strict=False should try
        # to get selections (but current code fails, so we also raise SelectionError)
        ref = reference.select_atoms(selection).atoms[1:]
        mobile = universe.select_atoms(selection)
        if strict:
            with pytest.raises(SelectionError):
                groups = align.get_matching_atoms(ref, mobile, strict=strict)
        else:
            with pytest.warns(SelectionWarning):
                with pytest.raises(SelectionError):
                    groups = align.get_matching_atoms(
                        ref, mobile, strict=strict
                    )

    @pytest.mark.parametrize("strict", (True, False))
    def test_nomatch_residues_raise_empty(
        self,
        universe,
        reference_small,
        strict,
        selection="protein and backbone",
    ):
        # one atom less and all residues different: will currently create
        # empty selections with strict=False, see also
        # https://gist.github.com/orbeckst/2686badcd15031e6c946baf9164a683d
        ref = reference_small.select_atoms(selection)
        mobile = universe.select_atoms(selection)
        if strict:
            with pytest.raises(SelectionError):
                groups = align.get_matching_atoms(ref, mobile, strict=strict)
        else:
            with pytest.warns(SelectionWarning):
                with pytest.raises(SelectionError):
                    groups = align.get_matching_atoms(
                        ref, mobile, strict=strict
                    )

    def test_toggle_atom_mismatch_default_error(self, universe, reference):
        selection = ("resname ALA and name CA", "resname ALA and name O")
        with pytest.raises(SelectionError):
            rmsd = align.alignto(universe, reference, select=selection)

    def test_toggle_atom_mismatch_kwarg_error(self, universe, reference):
        selection = ("resname ALA and name CA", "resname ALA and name O")
        with pytest.raises(SelectionError):
            rmsd = align.alignto(
                universe, reference, select=selection, match_atoms=True
            )

    def test_toggle_atom_nomatch(self, universe, reference):
        selection = ("resname ALA and name CA", "resname ALA and name O")
        rmsd = align.alignto(
            universe, reference, select=selection, match_atoms=False
        )
        assert rmsd[0] > 0.01

    def test_toggle_atom_nomatch_mismatch_atoms(self, universe, reference):
        # mismatching number of atoms, but same number of residues
        u = universe.select_atoms("resname ALA and name CA")
        u += universe.select_atoms("resname ALA and name O")[-1]
        ref = reference.select_atoms("resname ALA and name CA")
        with pytest.raises(SelectionError):
            align.alignto(u, ref, select="all", match_atoms=False)

    @pytest.mark.parametrize(
        "subselection, expectation",
        [
            ("resname ALA and name CA", does_not_raise()),
            (
                mda.Universe(PSF, DCD).select_atoms("resname ALA and name CA"),
                does_not_raise(),
            ),
            (1234, pytest.raises(TypeError)),
        ],
    )
    def test_subselection_alignto(
        self, universe, reference, subselection, expectation
    ):

        with expectation:
            rmsd = align.alignto(
                universe, reference, subselection=subselection
            )
            assert_allclose(rmsd[1], 0.0, rtol=0, atol=1.5e-9)

    def test_no_atom_masses(self, universe):
        # if no masses are present
        u = mda.Universe.empty(
            6, 2, atom_resindex=[0, 0, 0, 1, 1, 1], trajectory=True
        )
        with pytest.warns(SelectionWarning):
            align.get_matching_atoms(u.atoms, u.atoms)

    def test_one_universe_has_masses(self, universe):
        u = mda.Universe.empty(
            6, 2, atom_resindex=[0, 0, 0, 1, 1, 1], trajectory=True
        )
        ref = mda.Universe.empty(
            6, 2, atom_resindex=[0, 0, 0, 1, 1, 1], trajectory=True
        )
        ref.add_TopologyAttr("masses")
        with pytest.warns(SelectionWarning):
            align.get_matching_atoms(u.atoms, ref.atoms)


class TestAlign(object):
    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PSF, DCD)

    @staticmethod
    @pytest.fixture()
    def reference():
        return mda.Universe(PSF, DCD)

    def test_rmsd(self, universe, reference):
        universe.trajectory[0]  # ensure first frame
        bb = universe.select_atoms("backbone")
        first_frame = bb.positions
        universe.trajectory[-1]
        last_frame = bb.positions
        assert_allclose(
            rms.rmsd(first_frame, first_frame),
            0.0,
            rtol=0,
            atol=1.5e-5,
            err_msg="error: rmsd(X,X) should be 0",
        )
        # rmsd(A,B) = rmsd(B,A) should be exact but spurious failures in the
        # 9th decimal have been observed (see Issue 57 comment #1) so we relax
        # the test to 6 decimals.
        rmsd = rms.rmsd(first_frame, last_frame, superposition=True)
        assert_allclose(
            rms.rmsd(last_frame, first_frame, superposition=True),
            rmsd,
            rtol=0,
            atol=1.5e-6,
            err_msg="error: rmsd() is not symmetric",
        )
        assert_allclose(
            rmsd,
            6.820321761927005,
            rtol=0,
            atol=1.5e-5,
            err_msg="RMSD calculation between 1st and last AdK frame gave wrong answer",
        )
        # test masses as weights
        last_atoms_weight = universe.atoms.masses
        A = universe.trajectory[0]
        B = reference.trajectory[-1]
        rmsd = align.alignto(universe, reference, weights="mass")
        rmsd_sup_weight = rms.rmsd(
            A, B, weights=last_atoms_weight, center=True, superposition=True
        )
        assert_allclose(rmsd[1], rmsd_sup_weight, rtol=0, atol=1.5e-6)

    def test_rmsd_custom_mass_weights(self, universe, reference):
        last_atoms_weight = universe.atoms.masses
        A = universe.trajectory[0]
        B = reference.trajectory[-1]
        rmsd = align.alignto(
            universe, reference, weights=reference.atoms.masses
        )
        rmsd_sup_weight = rms.rmsd(
            A, B, weights=last_atoms_weight, center=True, superposition=True
        )
        assert_allclose(rmsd[1], rmsd_sup_weight, rtol=0, atol=1.5e-6)

    def test_rmsd_custom_weights(self, universe, reference):
        weights = np.zeros(universe.atoms.n_atoms)
        ca = universe.select_atoms("name CA")
        weights[ca.indices] = 1
        rmsd = align.alignto(universe, reference, select="name CA")
        rmsd_weights = align.alignto(universe, reference, weights=weights)
        assert_allclose(rmsd[1], rmsd_weights[1], rtol=0, atol=1.5e-6)

    def test_AlignTraj_outfile_default(self, universe, reference, tmpdir):
        with tmpdir.as_cwd():
            reference.trajectory[-1]
            x = align.AlignTraj(universe, reference)
            try:
                assert os.path.basename(x.filename) == "rmsfit_adk_dims.dcd"
            finally:
                x._writer.close()

    def test_AlignTraj_outfile_default_exists(
        self, universe, reference, tmpdir
    ):
        reference.trajectory[-1]
        outfile = str(tmpdir.join("align_test.dcd"))
        align.AlignTraj(universe, reference, filename=outfile).run()
        fitted = mda.Universe(PSF, outfile)

        # ensure default file exists
        with mda.Writer(
            str(tmpdir.join("rmsfit_align_test.dcd")),
            n_atoms=fitted.atoms.n_atoms,
        ) as w:
            w.write(fitted.atoms)

        with tmpdir.as_cwd():
            align.AlignTraj(fitted, reference)

            # we are careful now. The default does nothing
            with pytest.raises(IOError):
                align.AlignTraj(fitted, reference, force=False)

    def test_AlignTraj_step_works(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join("align_test.dcd"))
        # this shouldn't throw an exception
        align.AlignTraj(universe, reference, filename=outfile).run(step=10)

    def test_AlignTraj_deprecated_attribute(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join("align_test.dcd"))
        x = align.AlignTraj(universe, reference, filename=outfile).run(stop=2)

        wmsg = "The `rmsd` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(x.rmsd, x.results.rmsd)

    def test_AlignTraj(self, universe, reference, tmpdir):
        reference.trajectory[-1]
        outfile = str(tmpdir.join("align_test.dcd"))
        x = align.AlignTraj(universe, reference, filename=outfile).run()
        fitted = mda.Universe(PSF, outfile)

        assert_allclose(x.results.rmsd[0], 6.9290, rtol=0, atol=1.5e-3)
        assert_allclose(x.results.rmsd[-1], 5.2797e-07, rtol=0, atol=1.5e-3)

        # RMSD against the reference frame
        # calculated on Mac OS X x86 with MDA 0.7.2 r689
        # VMD: 6.9378711
        self._assert_rmsd(reference, fitted, 0, 6.929083044751061)
        self._assert_rmsd(reference, fitted, -1, 0.0)

    def test_AlignTraj_weighted(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join("align_test.dcd"))
        x = align.AlignTraj(
            universe, reference, filename=outfile, weights="mass"
        ).run()
        fitted = mda.Universe(PSF, outfile)
        assert_allclose(x.results.rmsd[0], 0, rtol=0, atol=1.5e-3)
        assert_allclose(x.results.rmsd[-1], 6.9033, rtol=0, atol=1.5e-3)

        self._assert_rmsd(
            reference, fitted, 0, 0.0, weights=universe.atoms.masses
        )
        self._assert_rmsd(
            reference,
            fitted,
            -1,
            6.929083032629219,
            weights=universe.atoms.masses,
        )

    def test_AlignTraj_custom_weights(self, universe, reference, tmpdir):
        weights = np.zeros(universe.atoms.n_atoms)
        ca = universe.select_atoms("name CA")
        weights[ca.indices] = 1

        outfile = str(tmpdir.join("align_test.dcd"))

        x = align.AlignTraj(
            universe, reference, filename=outfile, select="name CA"
        ).run()
        x_weights = align.AlignTraj(
            universe, reference, filename=outfile, weights=weights
        ).run()

        assert_allclose(
            x.results.rmsd, x_weights.results.rmsd, rtol=0, atol=1.5e-7
        )

    def test_AlignTraj_custom_mass_weights(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join("align_test.dcd"))
        x = align.AlignTraj(
            universe,
            reference,
            filename=outfile,
            weights=reference.atoms.masses,
        ).run()
        fitted = mda.Universe(PSF, outfile)
        assert_allclose(x.results.rmsd[0], 0, rtol=0, atol=1.5e-3)
        assert_allclose(x.results.rmsd[-1], 6.9033, rtol=0, atol=1.5e-3)

        self._assert_rmsd(
            reference, fitted, 0, 0.0, weights=universe.atoms.masses
        )
        self._assert_rmsd(
            reference,
            fitted,
            -1,
            6.929083032629219,
            weights=universe.atoms.masses,
        )

    def test_AlignTraj_partial_fit(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join("align_test.dcd"))
        # fitting on a partial selection should still write the whole topology
        align.AlignTraj(
            universe,
            reference,
            select="resid 1-20",
            filename=outfile,
            weights="mass",
        ).run()
        mda.Universe(PSF, outfile)

    def test_AlignTraj_in_memory(self, universe, reference, tmpdir):
        outfile = str(tmpdir.join("align_test.dcd"))
        reference.trajectory[-1]
        x = align.AlignTraj(
            universe, reference, filename=outfile, in_memory=True
        ).run()
        assert x.filename is None
        assert_allclose(x.results.rmsd[0], 6.9290, rtol=0, atol=1.5e-3)
        assert_allclose(x.results.rmsd[-1], 5.2797e-07, rtol=0, atol=1.5e-3)

        # check in memory trajectory
        self._assert_rmsd(reference, universe, 0, 6.929083044751061)
        self._assert_rmsd(reference, universe, -1, 0.0)

    def test_AlignTraj_writer_kwargs(self, universe, reference, tmpdir):
        # Issue 4564
        writer_kwargs = dict(precision=2)
        with tmpdir.as_cwd():
            aligner = align.AlignTraj(
                universe,
                reference,
                select="protein and name CA",
                filename="aligned_traj.xtc",
                writer_kwargs=writer_kwargs,
                in_memory=False,
            ).run()
            assert_equal(aligner._writer.precision, 2)

    def _assert_rmsd(self, reference, fitted, frame, desired, weights=None):
        fitted.trajectory[frame]
        rmsd = rms.rmsd(
            reference.atoms.positions,
            fitted.atoms.positions,
            superposition=True,
        )
        assert_allclose(
            rmsd,
            desired,
            rtol=0,
            atol=1.5e-5,
            err_msg="frame {0:d} of fit does not have "
            "expected RMSD".format(frame),
        )

    def test_alignto_checks_selections(self, universe, reference):
        """Testing that alignto() fails if selections do not
        match (Issue 143)"""
        u = universe

        def different_size():
            a = u.atoms[10:100]
            b = u.atoms[10:101]
            return align.alignto(a, b)

        with pytest.raises(SelectionError):
            different_size()

        def different_atoms():
            a = u.atoms[10:20]
            b = u.atoms[10:17] + u.atoms[18:21]
            return align.alignto(a, b)

        with pytest.raises(SelectionError):
            different_atoms()

    def test_alignto_partial_universe(self, universe, reference):
        u_bound = mda.Universe(ALIGN_BOUND)
        u_free = mda.Universe(ALIGN_UNBOUND)
        selection = "segid B"

        segB_bound = u_bound.select_atoms(selection)
        segB_free = u_free.select_atoms(selection)
        segB_free.translate(segB_bound.centroid() - segB_free.centroid())

        align.alignto(u_free, u_bound, select=selection)
        assert_allclose(
            segB_bound.positions, segB_free.positions, rtol=0, atol=1.5e-3
        )


def _get_aligned_average_positions(ref_files, ref, select="all", **kwargs):
    u = mda.Universe(*ref_files, in_memory=True)
    prealigner = align.AlignTraj(u, ref, select=select, **kwargs).run()
    ag = u.select_atoms(select)
    reference_coordinates = u.trajectory.timeseries(asel=ag).mean(axis=1)
    rmsd = sum(prealigner.results.rmsd / len(u.trajectory))
    return reference_coordinates, rmsd


class TestAverageStructure(object):

    ref_files = (PSF, DCD)

    @pytest.fixture
    def universe(self):
        return mda.Universe(*self.ref_files)

    @pytest.fixture
    def reference(self):
        return mda.Universe(PSF, CRD)

    def test_average_structure_deprecated_attrs(self, universe, reference):
        # Issue #3278 - remove in MDAnalysis 3.0.0
        avg = align.AverageStructure(universe, reference).run(stop=2)

        wmsg = "The `universe` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(
                avg.universe.atoms.positions,
                avg.results.universe.atoms.positions,
            )

        wmsg = "The `positions` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(avg.positions, avg.results.positions)

        wmsg = "The `rmsd` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert avg.rmsd == avg.results.rmsd

    def test_average_structure(self, universe, reference):
        ref, rmsd = _get_aligned_average_positions(self.ref_files, reference)
        avg = align.AverageStructure(universe, reference).run()
        assert_allclose(
            avg.results.universe.atoms.positions, ref, rtol=0, atol=1.5e-4
        )
        assert_allclose(avg.results.rmsd, rmsd, rtol=0, atol=1.5e-7)

    def test_average_structure_mass_weighted(self, universe, reference):
        ref, rmsd = _get_aligned_average_positions(
            self.ref_files, reference, weights="mass"
        )
        avg = align.AverageStructure(universe, reference, weights="mass").run()
        assert_allclose(
            avg.results.universe.atoms.positions, ref, rtol=0, atol=1.5e-4
        )
        assert_allclose(avg.results.rmsd, rmsd, rtol=0, atol=1.5e-7)

    def test_average_structure_select(self, universe, reference):
        select = "protein and name CA and resid 3-5"
        ref, rmsd = _get_aligned_average_positions(
            self.ref_files, reference, select=select
        )
        avg = align.AverageStructure(universe, reference, select=select).run()
        assert_allclose(
            avg.results.universe.atoms.positions, ref, rtol=0, atol=1.5e-4
        )
        assert_allclose(avg.results.rmsd, rmsd, rtol=0, atol=1.5e-7)

    def test_average_structure_no_ref(self, universe):
        ref, rmsd = _get_aligned_average_positions(self.ref_files, universe)
        avg = align.AverageStructure(universe).run()
        assert_allclose(
            avg.results.universe.atoms.positions, ref, rtol=0, atol=1.5e-4
        )
        assert_allclose(avg.results.rmsd, rmsd, rtol=0, atol=1.5e-7)

    def test_average_structure_no_msf(self, universe):
        avg = align.AverageStructure(universe).run()
        assert not hasattr(avg, "msf")

    def test_mismatch_atoms(self, universe):
        u = mda.Merge(universe.atoms[:10])
        with pytest.raises(SelectionError):
            align.AverageStructure(universe, u)

    def test_average_structure_ref_frame(self, universe):
        ref_frame = 3
        u = mda.Merge(universe.atoms)

        # change to ref_frame
        universe.trajectory[ref_frame]
        u.load_new(universe.atoms.positions)

        # back to start
        universe.trajectory[0]
        ref, rmsd = _get_aligned_average_positions(self.ref_files, u)
        avg = align.AverageStructure(universe, ref_frame=ref_frame).run()
        assert_allclose(
            avg.results.universe.atoms.positions, ref, rtol=0, atol=1.5e-4
        )
        assert_allclose(avg.results.rmsd, rmsd, rtol=0, atol=1.5e-7)

    def test_average_structure_in_memory(self, universe):
        avg = align.AverageStructure(universe, in_memory=True).run()
        reference_coordinates = universe.trajectory.timeseries().mean(axis=1)
        assert_allclose(
            avg.results.universe.atoms.positions,
            reference_coordinates,
            rtol=0,
            atol=1.5e-4,
        )
        assert avg.filename is None


class TestAlignmentProcessing:
    seq = FASTA
    error_msg = "selection string has unexpected length"

    @pytest.mark.skipif(HAS_BIOPYTHON, reason="biopython is installed")
    def test_importerror_biopython(self):
        errmsg = "The `fasta2select` method requires an installation"
        with pytest.raises(ImportError, match=errmsg):
            _ = align.fasta2select(self.seq, is_aligned=True)

    @pytest.mark.skipif(not HAS_BIOPYTHON, reason="requires biopython")
    def test_fasta2select_aligned(self):
        """test align.fasta2select() on aligned FASTA (Issue 112)"""
        sel = align.fasta2select(self.seq, is_aligned=True)
        # length of the output strings, not residues or anything real...
        assert len(sel["reference"]) == 30623, self.error_msg
        assert len(sel["mobile"]) == 30623, self.error_msg

    @pytest.mark.skipif(
        executable_not_found("clustalw2") or not HAS_BIOPYTHON,
        reason="Test skipped because clustalw2 executable not found",
    )
    def test_fasta2select_file(self, tmpdir):
        """test align.fasta2select() on a non-aligned FASTA with default
        filenames"""
        with tmpdir.as_cwd():
            sel = align.fasta2select(
                self.seq, is_aligned=False, alnfilename=None, treefilename=None
            )
            assert len(sel["reference"]) == 23080, self.error_msg
            assert len(sel["mobile"]) == 23090, self.error_msg

    @pytest.mark.skipif(
        executable_not_found("clustalw2") or not HAS_BIOPYTHON,
        reason="Test skipped because clustalw2 executable not found",
    )
    def test_fasta2select_ClustalW(self, tmpdir):
        """MDAnalysis.analysis.align: test fasta2select() with ClustalW
        (Issue 113)"""
        alnfile = str(tmpdir.join("alignmentprocessing.aln"))
        treefile = str(tmpdir.join("alignmentprocessing.dnd"))
        sel = align.fasta2select(
            self.seq,
            is_aligned=False,
            alnfilename=alnfile,
            treefilename=treefile,
        )
        # numbers computed from alignment with clustalw 2.1 on Mac OS X
        # [orbeckst] length of the output strings, not residues or anything
        # real...
        assert len(sel["reference"]) == 23080, self.error_msg
        assert len(sel["mobile"]) == 23090, self.error_msg

    @pytest.mark.skipif(not HAS_BIOPYTHON, reason="requires biopython")
    def test_fasta2select_resids(self, tmpdir):
        """test align.fasta2select() when resids provided (Issue #3124)"""
        resids = [x for x in range(705)]
        sel = align.fasta2select(
            self.seq, is_aligned=True, ref_resids=resids, target_resids=resids
        )
        # length of the output strings, not residues or anything real...
        assert len(sel["reference"]) == 30621, self.error_msg
        assert len(sel["mobile"]) == 30621, self.error_msg


class TestSequenceAlignmentFunction:
    # remove 3.0

    @staticmethod
    @pytest.fixture
    def atomgroups():
        universe = mda.Universe(PSF)
        reference = universe.atoms
        mobile = universe.select_atoms("resid 122-159")
        return reference, mobile

    @pytest.mark.skipif(HAS_BIOPYTHON, reason="biopython installed")
    def test_biopython_import_error(self, atomgroups):
        ref, mob = atomgroups
        errmsg = "The `sequence_alignment` method requires an installation of"
        with pytest.raises(ImportError, match=errmsg):
            align.sequence_alignment(mob, ref)

    @pytest.mark.skipif(not HAS_BIOPYTHON, reason="requires biopython")
    @pytest.mark.filterwarnings("ignore:`sequence_alignment` is deprecated!")
    def test_sequence_alignment(self, atomgroups):
        reference, mobile = atomgroups
        aln = align.sequence_alignment(mobile, reference)

        assert len(aln) == 5, "return value has wrong tuple size"

        seqA, seqB, score, begin, end = aln
        assert_equal(
            seqA,
            reference.residues.sequence(format="string"),
            err_msg="reference sequence mismatch",
        )
        assert (
            mobile.residues.sequence(format="string") in seqB
        ), "mobile sequence mismatch"
        assert score == pytest.approx(54.6)
        assert_array_equal([begin, end], [0, reference.n_residues])

    @pytest.mark.skipif(not HAS_BIOPYTHON, reason="requires biopython")
    def test_sequence_alignment_deprecation(self, atomgroups):
        reference, mobile = atomgroups
        wmsg = (
            "`sequence_alignment` is deprecated!\n"
            "`sequence_alignment` will be removed in release 3.0."
        )
        with pytest.warns(DeprecationWarning, match=wmsg):
            align.sequence_alignment(mobile, reference)


class TestIterativeAverage(object):
    @pytest.fixture()
    def mobile(self):
        u = mda.Universe(PSF, DCD)
        return u

    @pytest.fixture()
    def reference(self):
        u = mda.Universe(PSF, DCD)
        return u

    def test_iterative_average_default(self, mobile):
        res = align.iterative_average(mobile, select="bynum 1:10")
        assert_allclose(
            res.results.positions,
            [
                [11.93075595, 8.6729893, -10.49887605],
                [12.60587898, 7.91673117, -10.73327464],
                [12.45662411, 9.51900517, -10.35551193],
                [11.27452274, 8.83003843, -11.2619057],
                [11.25808119, 8.26794477, -9.23340715],
                [12.02767222, 7.95332228, -8.57249317],
                [10.54679871, 9.49505306, -8.61215292],
                [9.99500556, 9.16624224, -7.75231192],
                [9.83897407, 9.93134598, -9.29541129],
                [11.45760169, 10.5857071, -8.13037669],
            ],
            atol=1e-5,
        )

    def test_iterative_average_eps_high(self, mobile):
        res = align.iterative_average(mobile, select="bynum 1:10", eps=1e-5)
        assert_allclose(
            res.results.positions,
            [
                [11.93075595, 8.6729893, -10.49887605],
                [12.60587898, 7.91673117, -10.73327464],
                [12.45662411, 9.51900517, -10.35551193],
                [11.27452274, 8.83003843, -11.2619057],
                [11.25808119, 8.26794477, -9.23340715],
                [12.02767222, 7.95332228, -8.57249317],
                [10.54679871, 9.49505306, -8.61215292],
                [9.99500556, 9.16624224, -7.75231192],
                [9.83897407, 9.93134598, -9.29541129],
                [11.45760169, 10.5857071, -8.13037669],
            ],
            atol=1e-5,
        )

    def test_iterative_average_weights_mass(self, mobile, reference):
        res = align.iterative_average(
            mobile, reference, select="bynum 1:10", niter=10, weights="mass"
        )
        assert_allclose(
            res.results.positions,
            [
                [11.96438784, 8.85426235, -10.24735737],
                [12.75920431, 8.27294545, -10.54295766],
                [12.3285704, 9.72083717, -9.9419435],
                [11.33941507, 9.03249423, -11.01306158],
                [11.30988499, 8.14958885, -9.1205501],
                [12.09108655, 7.85155906, -8.46681943],
                [10.37499697, 9.13535837, -8.3732586],
                [9.83883314, 8.57939098, -7.6195549],
                [9.64405257, 9.55924307, -9.04315991],
                [11.0678934, 10.27798773, -7.64881842],
            ],
            atol=1e-5,
        )

    def test_iterative_average_convergence_failure(self, mobile, reference):
        with pytest.raises(RuntimeError):
            _ = align.iterative_average(mobile, reference, niter=1, eps=0)

    def test_iterative_average_convergence_verbose(self, mobile, reference):
        _ = align.iterative_average(mobile, select="bynum 1:10", verbose=True)


def test_alignto_reorder_atomgroups():
    # Issue 2977
    u = mda.Universe(PDB_helix)
    mobile = u.atoms[:4]
    ref = u.atoms[[3, 2, 1, 0]]
    rmsd = align.alignto(mobile, ref, select="bynum 1-4")
    assert_allclose(rmsd, (0.0, 0.0))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
PQR file format --- :mod:`MDAnalysis.coordinates.PQR`
=====================================================

Read atoms with charges from a PQR_ file (as written by PDB2PQR_). The
following is adopted from the description of the PQR_ format as used by APBS_:

*MDAnalysis* reads very loosely-formatted PQR files: all fields are
**whitespace-delimited** rather than the strict column formatting mandated
by the PDB_ format. This more liberal formatting allows coordinates
which are larger/smaller than ±999 Å.

MDAnalysis reads data on a per-line basis from PQR files using the following format::

   recordName serial atomName residueName chainID residueNumber X Y Z charge radius

If this fails it is assumed that the *chainID* was omitted and the shorter
format is read::

   recordName serial atomName residueName residueNumber X Y Z charge radius

Anything else will raise a :exc:`ValueError`.

The whitespace is the most important feature of this format: fields
*must* be separated by at least one space or tab character. The fields
are:

*recordName*
    A string which specifies the type of PQR entry and should either be ATOM or
    HETATM.
*serial*
    An integer which provides the atom index (but note that MDAnalysis renumbers
    atoms so one cannot rely on the *serial*)
*atomName*
    A string which provides the atom name.
*residueName*
    A string which provides the residue name.
*chainID*
    An optional string which provides the chain ID of the atom.
*residueNumber*
    An integer which provides the residue index.
*X Y Z*
    Three floats which provide the atomic coordiantes.
*charge*
    A float which provides the atomic charge (in electrons).
*radius*
    A float which provides the atomic radius (in Å).

Clearly, this format can deviate wildly from PDB_ due to the use of whitespaces
rather than specific column widths and alignments. This deviation can be
particularly significant when large coordinate values are used.

Output should look like this (although the only real requirement is
*whitespace* separation between *all* entries). The chainID is optional
and can be omitted::

  ATOM      1  N    MET     1     -11.921   26.307   10.410 -0.3000 1.8500
  ATOM     36  NH1  ARG     2      -6.545   25.499    3.854 -0.8000 1.8500
  ATOM     37 HH11  ARG     2      -6.042   25.480    4.723  0.4600 0.2245


.. Warning:: Fields *must be white-space separated* or data are read
             incorrectly. PDB formatted files are *not* guaranteed to be
             white-space separated so extra care should be taken when quickly
             converting PDB files into PQR files using simple scripts.

For example, PQR files created with PDB2PQR_ and the `--whitespace`
option are guaranteed to conform to the above format::

   pdb2pqr --ff=charmm --whitespace 4ake.pdb 4ake.pqr


Notes
-----
The PQR format does not provide a means by which to provide box information.
In all cases the `dimensions` attribute will be set to `None`.


.. _PQR:     https://apbs.readthedocs.io/en/latest/formats/pqr.html
.. _APBS:    https://apbs.readthedocs.io/en/latest/
.. _PDB2PQR: https://pdb2pqr.readthedocs.io/en/latest/
.. _PDB:     http://www.wwpdb.org/documentation/file-format
"""
import itertools
import numpy as np
import warnings

from ..lib import util
from . import base


class PQRReader(base.SingleFrameReaderBase):
    """Read a PQR_ file into MDAnalysis.

    .. _PQR:
        https://apbs.readthedocs.io/en/latest/formats/pqr.html

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based
    """

    format = "PQR"
    units = {"time": None, "length": "Angstrom"}

    # how to slice fields[x:y] to grab coordinates
    _SLICE_INDICES = {
        "ORIGINAL": (-5, -2),
        "NO_CHAINID": (-5, -2),
        "GROMACS": (-6, -3),
    }

    def _read_first_frame(self):
        from ..topology.PQRParser import PQRParser

        flavour = None

        coords = []
        with util.openany(self.filename) as pqrfile:
            for line in pqrfile:
                if line.startswith(("ATOM", "HETATM")):
                    if flavour is None:
                        flavour = PQRParser.guess_flavour(line)
                        x, y = self._SLICE_INDICES[flavour]

                    fields = line.split()
                    # convert all entries at the end once for optimal speed
                    coords.append(fields[x:y])

        self.n_atoms = len(coords)
        self.ts = self._Timestep.from_coordinates(coords, **self._ts_kwargs)
        self.ts.frame = 0  # 0-based frame number
        if self.convert_units:
            # in-place !
            self.convert_pos_from_native(self.ts._pos)

    def Writer(self, filename, **kwargs):
        """Returns a PQRWriter for *filename*.

        Parameters
        ----------
        filename : str
            filename of the output PQR file

        Returns
        -------
        :class:`PQRWriter`
        """
        return PQRWriter(filename, **kwargs)


class PQRWriter(base.WriterBase):
    """Write a single coordinate frame in whitespace-separated PQR format.

    Charges ("Q") are taken from the
    :attr:`MDAnalysis.core.groups.Atom.charge` attribute while
    radii are obtaine from the
    :attr:`MDAnalysis.core.groups.Atom.radius` attribute.

    * If the segid is 'SYSTEM' then it will be set to the empty
      string. Otherwise the first letter will be used as the chain ID.
    * The serial number always starts at 1 and increments sequentially
      for the atoms.

    The output format is similar to ``pdb2pqr --whitespace``.


    .. versionadded:: 0.9.0
    .. versionchanged:: 2.6.0
       Files are now written in `wt` mode, and keep extensions, allowing
       for files to be written under compressed formats
    """

    format = "PQR"
    units = {"time": None, "length": "Angstrom"}

    # serial, atomName, residueName, chainID, residueNumber, XYZ, charge, radius
    fmt_ATOM = (
        "ATOM {serial:6d} {name:<4}  {resname:<3} {chainid:1.1}"
        " {resid:4d}   {pos[0]:-8.3f} {pos[1]:-8.3f}"
        " {pos[2]:-8.3f} {charge:-7.4f} {radius:6.4f}\n"
    )
    fmt_remark = "REMARK   {0} {1}\n"

    def __init__(self, filename, convert_units=True, **kwargs):
        """Set up a PQRWriter with full whitespace separation.

        Parameters
        ----------
        filename : str
             output filename
        convert_units: bool (optional)
            units are converted to the MDAnalysis base format; [``True``]
        remarks : str (optional)
             remark lines (list of strings) or single string to be added to the
             top of the PQR file
        """
        self.filename = util.filename(filename, ext="pqr", keep=True)
        self.convert_units = (
            convert_units  # convert length and time to base units
        )
        self.remarks = kwargs.pop("remarks", "PQR file written by MDAnalysis")

    def write(self, selection, frame=None):
        """Write selection at current trajectory frame to file.

        Parameters
        ----------
        selection : AtomGroup or Universe
            MDAnalysis AtomGroup or Universe
        frame : int (optional)
            optionally move to frame index `frame`; by default, write the
            current frame


        .. versionchanged:: 0.11.0
           Frames now 0-based instead of 1-based

        """
        # write() method that complies with the Trajectory API
        try:
            u = selection.universe
        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None

        if frame is not None:
            u.trajectory[frame]  # advance to frame
        else:
            try:
                frame = u.trajectory.ts.frame
            except AttributeError:
                frame = 0  # should catch cases when we are analyzing a single frame(?)

        atoms = selection.atoms  # make sure to use atoms (Issue 46)
        coordinates = (
            atoms.positions
        )  # can write from selection == Universe (Issue 49)
        if self.convert_units:
            self.convert_pos_to_native(
                coordinates
            )  # inplace because coordinates is already a copy

        # Check atom attributes
        # required:
        # - name
        # - resname
        # - chainid
        # - resid
        # - position
        # - charge
        # - radius
        attrs = {}
        missing_topology = []
        for attr, dflt in (
            ("names", itertools.cycle(("X",))),
            ("resnames", itertools.cycle(("UNK",))),
            ("resids", itertools.cycle((1,))),
            ("charges", itertools.cycle((0.0,))),
            ("radii", itertools.cycle((1.0,))),
        ):
            try:
                attrs[attr] = getattr(atoms, attr)
            except AttributeError:
                attrs[attr] = dflt
                missing_topology.append(attr)

        # chainids require special handling
        # try chainids, then segids
        # if neither, use ' '
        # if 'SYSTEM', use ' '
        try:
            attrs["chainids"] = atoms.chainids
        except AttributeError:
            try:
                attrs["chainids"] = atoms.segids
            except AttributeError:
                pass
        if not "chainids" in attrs or all(attrs["chainids"] == "SYSTEM"):
            attrs["chainids"] = itertools.cycle((" ",))

        if "charges" in missing_topology:
            total_charge = 0.0
        else:
            total_charge = atoms.total_charge()

        if missing_topology:
            warnings.warn(
                "Supplied AtomGroup was missing the following attributes: "
                "{miss}. These will be written with default values. "
                "".format(miss=", ".join(missing_topology))
            )

        with util.openany(self.filename, "wt") as pqrfile:
            # Header / Remarks
            # The *remarknumber* is typically 1 but :program:`pdb2pgr`
            # also uses 6 for the total charge and 5 for warnings.
            for rem in util.asiterable(self.remarks):
                pqrfile.write(self.fmt_remark.format(rem, 1))
            pqrfile.write(
                self.fmt_remark.format(
                    "Input: frame {0} of {1}".format(
                        frame, u.trajectory.filename
                    ),
                    5,
                )
            )
            pqrfile.write(
                self.fmt_remark.format(
                    "total charge: {0:+8.4f} e".format(total_charge), 6
                )
            )

            # Atom descriptions and coords
            for atom_index, (
                pos,
                name,
                resname,
                chainid,
                resid,
                charge,
                radius,
            ) in enumerate(
                zip(
                    coordinates,
                    attrs["names"],
                    attrs["resnames"],
                    attrs["chainids"],
                    attrs["resids"],
                    attrs["charges"],
                    attrs["radii"],
                ),
                start=1,
            ):
                # pad so that only 4-letter atoms are left-aligned
                name = " " + name if len(name) < 4 else name

                pqrfile.write(
                    self.fmt_ATOM.format(
                        serial=atom_index,
                        name=name,
                        resname=resname,
                        chainid=chainid,
                        resid=resid,
                        pos=pos,
                        charge=charge,
                        radius=radius,
                    )
                )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np

from MDAnalysisTests import datafiles
from MDAnalysisTests.datafiles import (
    PDB_small,
    PDB,
    LAMMPSdata,
    LAMMPSdata2,
    LAMMPSdcd2,
    LAMMPSdata_mini,
    LAMMPSdata_additional_columns,
    PSF_TRICLINIC,
    DCD_TRICLINIC,
    PSF_NAMD_TRICLINIC,
    DCD_NAMD_TRICLINIC,
)


class RefAdKSmall(object):
    """Mixin class to provide comparison numbers.

    Based on small PDB with AdK (:data:`PDB_small`).

    .. Note::

       All distances must be in ANGSTROEM as this is the MDAnalysis
       default unit. All readers must return Angstroem by default.
    """

    filename = datafiles.PDB_small
    ref_coordinates = {
        # G11:CA, copied frm adk_open.pdb
        "A10CA": np.array([-1.198, 7.937, 22.654]),
    }
    ref_distances = {"endtoend": 11.016959}
    ref_E151HA2_index = 2314
    ref_n_atoms = 3341
    ref_charmm_totalcharge = -4.0
    ref_charmm_Hcharges = [0.33] + 203 * [0.31]
    ref_charmm_ArgCAcharges = 13 * [0.07]
    ref_charmm_ProNcharges = 10 * [-0.29]
    ref_unitcell = np.array(
        [80.017, 80.017, 80.017, 60.0, 60.0, 90.0], dtype=np.float32
    )
    ref_volume = 0.0


class RefAdK(object):
    """Mixin class to provide comparison numbers.

    Based on PDB/GRO with AdK in water + Na+ (:data:`PDB`).

    .. Note::

       All distances must be in ANGSTROEM as this is the MDAnalysis
       default unit. All readers must return Angstroem by default.
    """

    filename = datafiles.PDB
    ref_coordinates = {
        # Angstroem as MDAnalysis unit!!
        "A10CA": np.array([62.97600174, 62.08800125, 20.2329998]),
    }
    ref_distances = {"endtoend": 9.3513174}
    ref_E151HA2_index = 2314
    ref_n_atoms = 47681
    ref_Na_sel_size = 4
    # CRYST1 80.017   80.017   80.017  60.00  60.00  90.00
    ref_unitcell = np.array(
        [80.017, 80.017, 80.017, 60.0, 60.0, 90.0], dtype=np.float32
    )
    # ref_volume = 362270.0  # computed with Gromacs ## NOT EXACT!
    ref_volume = 362269.520669292


class Ref2r9r(object):
    """Mixin class to provide comparison numbers.

    Based on S6 helices of chimeric Kv channel

    .. Note::

       All distances must be in ANGSTROEM as this is the MDAnalysis
       default unit. All readers must return Angstroem by default.
    """

    ref_n_atoms = 1284
    ref_sum_centre_of_geometry = -98.24146
    ref_n_frames = 10


class RefACHE(object):
    """Mixin class to provide comparison numbers.

    ACHE peptide

    # COM check in VMD::

        set p [atomselect top "not water"]
        set total {0 0 0};
        for {set i 0} {$i < 11} {incr i} {
           $p frame $i; set total [vecadd $total [measure center $p]]}

        puts [vecsum $total]
        # 472.2592159509659

    """

    ref_n_atoms = 252
    ref_proteinatoms = ref_n_atoms
    ref_sum_centre_of_geometry = 472.2592159509659  # 430.44807815551758
    ref_n_frames = 11
    ref_periodic = False


class RefCappedAla(object):
    """Mixin class to provide comparison numbers.

    Capped Ala in water

    # COM check in VMD (load trajectory as *AMBER with periodic box*!)::

        set p [atomselect top "not water"]
        set total {0 0 0};
        for {set i 0} {$i < 11} {incr i} {
           $p frame $i; set total [vecadd $total [measure center $p]]}

        puts [vecsum $total]
        # 686.276834487915

    """

    ref_n_atoms = 5071
    ref_proteinatoms = 22
    ref_sum_centre_of_geometry = 686.276834487915
    ref_n_frames = 11
    ref_periodic = True


class RefVGV(object):
    """Mixin class to provide comparison numbers.

    Computed from bala.trj::

      w = MDAnalysis.Universe(PRMncdf, TRJncdf)
      ref_n_atoms = len(w.atoms) ref_proteinatoms = len(w.select_atoms("protein"))
      ref_sum_centre_of_geometry = np.sum([protein.center_of_geometry()
                                           for ts in w.trajectory])
    """

    topology = datafiles.PRMncdf
    filename = datafiles.NCDF
    ref_n_atoms = 2661
    ref_proteinatoms = 50
    ref_sum_centre_of_geometry = 1552.9125
    ref_n_frames = 30
    ref_periodic = True


class RefTZ2(object):
    """Reference values for the cpptraj testcase tz2.truncoct.nc

    Used under the GPL v3.
    """

    topology = datafiles.PRM7
    filename = datafiles.NCDFtruncoct
    ref_n_atoms = 5827
    ref_proteinatoms = 217
    ref_sum_centre_of_geometry = -68.575745
    ref_n_frames = 10
    ref_periodic = True


class RefTRZ(object):
    #    ref_coordinates = {}
    #    ref_distances = {'endtoend': }
    ref_n_atoms = 8184
    ref_dimensions = np.array(
        [
            55.422830581665039,
            55.422830581665039,
            55.422830581665039,
            90.0,
            90.0,
            90.0,
        ],
        dtype=np.float32,
    )
    ref_volume = 170241.762765
    ref_n_frames = 6
    ref_coordinates = np.array(
        [72.3163681, -130.31130981, 19.97969055], dtype=np.float32
    )
    ref_velocities = np.array(
        [[14.83297443, 18.02611542, 6.07733774]], dtype=np.float32
    )
    ref_delta = 0.001
    ref_time = 0.01
    ref_title = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ12345678901234"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ12345678901234"
    )


class RefLAMMPSData(object):
    filename = LAMMPSdata
    n_atoms = 18364
    pos_atom1 = np.array(
        [11.89985657, 48.4455719, 19.09719849], dtype=np.float32
    )
    vel_atom1 = np.array(
        [-5667.59300232, 7913.80977631, -3007.79533386], dtype=np.float32
    )
    dimensions = np.array(
        [55.42282867, 55.42282867, 55.42282867, 90.0, 90.0, 90.0],
        dtype=np.float32,
    )


class RefLAMMPSDataDCD(object):
    format = "LAMMPS"
    topology = LAMMPSdata2
    trajectory = LAMMPSdcd2
    n_atoms = 12421
    n_frames = 5
    dt = 0.5  # ps per frame
    mean_dimensions = np.array(
        [50.66186142, 47.18824387, 52.33762741, 90.0, 90.0, 90.0],
        dtype=np.float32,
    )


class RefLAMMPSDataMini(object):
    filename = LAMMPSdata_mini
    n_atoms = 1
    pos_atom1 = np.array(
        [11.89985657, 48.4455719, 19.09719849], dtype=np.float32
    )
    vel_atom1 = np.array(
        [-5667.59300232, 7913.80977631, -3007.79533386], dtype=np.float32
    )
    dimensions = np.array(
        [60.0, 50.0, 30.0, 90.0, 90.0, 90.0], dtype=np.float32
    )


class RefLAMMPSDataAdditionalColumns(object):
    q = np.array(
        [
            2.58855e-03,
            6.91952e-05,
            1.05548e-02,
            4.20319e-03,
            9.19172e-03,
            4.79777e-03,
            6.36864e-04,
            5.87125e-03,
            -2.18125e-03,
            6.88910e-03,
        ]
    )
    p = np.array(5 * [1.1, 1.2])



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


from libc.stdlib cimport free
from libc.stdint cimport uintptr_t
from libc.stdio cimport SEEK_SET, SEEK_CUR, SEEK_END
import cython

cimport numpy as cnp

cnp.import_array()


# Tell cython about the off_t type. It doesn't need to match exactly what is
# defined since we don't expose it to python but the cython compiler needs to
# know about it.
cdef extern from 'sys/types.h':
    ctypedef int off_t

ctypedef int fio_fd
ctypedef off_t fio_size_t

cdef extern from 'include/fastio.h':
    int fio_open(const char * filename, int mode, fio_fd * fd)
    int fio_fclose(fio_fd fd)
    fio_size_t fio_ftell(fio_fd fd)
    fio_size_t fio_fseek(fio_fd fd, fio_size_t offset, int whence)

cdef extern from 'include/readdcd.h':
    int read_dcdheader(fio_fd fd, int * natoms, int * nsets, int * istart,
                       int * nsavc, double * delta, int * nfixed, int ** freeind,
                       float ** fixedcoords, int * reverse_endian, int * charmm,
                       char ** remarks, int * len_remarks)
    void close_dcd_read(int * freeind, float * fixedcoords)
    int read_dcdstep(fio_fd fd, int natoms, float * X, float * Y, float * Z,
                     double * unitcell, int num_fixed,
                     int first, int * indexes, float * fixedcoords,
                     int reverse_endian, int charmm)
    int read_dcdsubset(fio_fd fd, int natoms, int lowerb, int upperb,
                       float * X, float * Y, float * Z,
                       double * unitcell, int num_fixed,
                       int first, int * indexes, float * fixedcoords,
                       int reverse_endian, int charmm)
    int write_dcdheader(fio_fd fd, const char * remarks, int natoms,
                        int istart, int nsavc, double delta, int with_unitcell,
                        int charmm)
    int write_dcdstep(fio_fd fd, int curframe, int curstep,
                      int natoms, const float * x, const float * y, const float * z,
                      const double * unitcell, int charmm)

cdef class DCDFile:
    """DCDFile(fname, mode='r')

    File like wrapper for DCD files

    This class can be similar to the normal file objects in python. The read()
    function will return a frame and all information in it instead of a single
    line. Additionally the context-manager protocol is supported as well.
    """
    # DCD file pointer
    cdef fio_fd fp
    # File name
    cdef readonly fname
    # Starting timestep of dcd file
    cdef int istart
    # Timesteps between dcd saves
    cdef int nsavc
    # Trajectory timestep
    cdef double delta
    # Number of atoms
    cdef int natoms
    # Number of fixed atoms
    cdef int nfixed
    # Free indices
    cdef int * freeind
    # Fixed coordinates
    cdef float * fixedcoords
    # Are we reverse endian?
    cdef int reverse_endian
    # Is the DCD file CHARMM style
    cdef int charmm
    # Is the file periodic
    cdef readonly int is_periodic
    # String data in the file
    cdef remarks
    # File mode
    cdef str mode
    # Number of dimensions in 
    cdef readonly int ndims
    # The number of frames
    cdef readonly int n_frames
    # Flag to indicate if header has been read
    cdef bint b_read_header
    # The current DCD frame
    cdef int current_frame
    # size of the first DCD frame
    cdef readonly fio_size_t _firstframesize
    # Size of a DCD frame 
    cdef readonly fio_size_t _framesize
    # Size of the DCD header
    cdef readonly fio_size_t _header_size
    # Is the file open?
    cdef int is_open
    # Have we reached the end of the file
    cdef int reached_eof
    # Have we written the header?
    cdef int wrote_header
    # Are buffers set up?
    cdef int _buffers_setup


    # buffer for reading coordinates
    cdef cnp.ndarray _coordinate_buffer
    # buffer for reading unitcell     
    cdef cnp.ndarray _unitcell_buffer

    # fortran contiguious memoryviews of the buffers to pass to the C code
    cdef float[::1] xview
    cdef float[::1] yview
    cdef float[::1] zview
    cdef double[::1] unitcellview

    cdef void _setup_buffers(self)

    cdef void _read_header(self)
    # Estimate the number of frames
    cdef int _estimate_n_frames(self)
    # Helper to read current DCD frame
    cdef int c_readframes_helper(self, float[::1] x,
                                 float[::1] y, float[::1] z,
                                 double[::1] unitcell, int first_frame)

# Helper in readframes to copy given a specific memory layout
cdef void copy_in_order(float[:, :] source, float[:, :, :] target, int order, int index)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal
import pytest

import MDAnalysis as mda
from MDAnalysis import NoDataError
from MDAnalysisTests.core.util import UnWrapUniverse, assert_in_box
from MDAnalysisTests.datafiles import TRZ_psf, TRZ


class TestWrap(object):
    """Tests the functionality of *Group.wrap() using the UnWrapUniverse,
    which is specifically designed for wrapping and unwrapping tests.
    """

    precision = 5

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    @pytest.mark.parametrize("is_triclinic", (False, True))
    def test_wrap_pass(self, level, compound, center, is_triclinic):
        # get a pristine test universe:
        u = UnWrapUniverse(is_triclinic=is_triclinic)
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        # store original positions:
        orig_pos = group.atoms.positions
        # get the expected result:
        ref_wrapped_pos = u.wrapped_coords(compound, center)
        # first, do the wrapping out-of-place:
        wrapped_pos = group.wrap(
            compound=compound, center=center, inplace=False
        )
        # check for correct result:
        assert_almost_equal(
            wrapped_pos, ref_wrapped_pos, decimal=self.precision
        )
        # make sure atom positions are unchanged:
        assert_array_equal(group.atoms.positions, orig_pos)
        # now, do the wrapping inplace:
        wrapped_pos2 = group.wrap(
            compound=compound, center=center, inplace=True
        )
        # check that result is the same as for out-of-place computation:
        assert_array_equal(wrapped_pos, wrapped_pos2)
        # check that wrapped positions are applied:
        assert_array_equal(group.atoms.positions, wrapped_pos)
        # check that nobody messed with the reference positions,
        # centers of compounds must lie within the primary unit cell:
        if compound == "atoms":
            assert_in_box(group.atoms.positions, group.dimensions)
        elif center == "com":
            compos = group.atoms.center_of_mass(wrap=False, compound=compound)
            assert_in_box(compos, group.dimensions)
        else:
            cogpos = group.atoms.center_of_geometry(
                wrap=False, compound=compound
            )
            assert_in_box(cogpos, group.dimensions)

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    @pytest.mark.parametrize("is_triclinic", (False, True))
    def test_unwrap_wrap_cycle(self, level, compound, center, is_triclinic):
        # get a pristine test universe:
        u = UnWrapUniverse(is_triclinic=is_triclinic)
        # set wrapped reference coordinates:
        u.atoms.positions = u.wrapped_coords("atoms", "com")
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        # unwrap:
        group.unwrap()
        # store original unwrapped positions:
        orig_unwrapped_pos = group.atoms.positions
        # wrap:
        group.wrap(compound=compound, center=center, inplace=True)
        # unwrap again:
        group.unwrap()
        # make sure unwrapped atom positions are as before:
        assert_almost_equal(
            group.atoms.positions, orig_unwrapped_pos, decimal=self.precision
        )

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    @pytest.mark.parametrize("is_triclinic", (False, True))
    def test_wrap_partial_compound(
        self, level, compound, center, is_triclinic
    ):
        # get a pristine test universe:
        u = UnWrapUniverse(is_triclinic=is_triclinic)
        group = u.atoms
        ref_wrapped_pos = u.wrapped_coords(compound, center)
        # select topology level with one missing item and get expected result:
        if level == "atoms":
            # first atom of molecule 12 missing
            missing = group[[-8]]
            group = group[:-8] + group[-7:]
            ref_wrapped_pos = np.concatenate(
                [ref_wrapped_pos[:-8], ref_wrapped_pos[-7:]]
            )
        elif level == "residues":
            group = group.residues
            # first residue of molecule 12 missing
            missing = group[-2]
            group = group[:-2] + group[[-1]]
            ref_wrapped_pos = np.concatenate(
                [ref_wrapped_pos[:-8], ref_wrapped_pos[-4:]]
            )
        elif level == "segments":
            group = group.segments
            # molecule 12 missing
            missing = group[-1]
            group = group[:-1]
            ref_wrapped_pos = ref_wrapped_pos[:-8]
        # store original positions of the missing item:
        orig_pos = missing.atoms.positions
        # first, do the wrapping out-of-place:
        group.wrap(compound=compound, center=center, inplace=True)
        # check for correct result:
        assert_almost_equal(
            group.atoms.positions, ref_wrapped_pos, decimal=self.precision
        )
        # make sure the positions of the missing item are unchanged:
        assert_array_equal(missing.atoms.positions, orig_pos)

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    @pytest.mark.parametrize("is_triclinic", (False, True))
    def test_wrap_empty_group(self, level, compound, center, is_triclinic):
        # get a pristine test universe:
        u = UnWrapUniverse(is_triclinic=is_triclinic)
        group = u.atoms[[]]
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        group.wrap(compound=compound, center=center, inplace=True)
        # check for correct (empty) result:
        assert_array_equal(
            group.atoms.positions, np.empty((0, 3), dtype=np.float32)
        )

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    @pytest.mark.parametrize("is_triclinic", (False, True))
    def test_wrap_duplicates(self, level, compound, center, is_triclinic):
        # get a pristine test universe:
        u = UnWrapUniverse(is_triclinic=is_triclinic)
        # select molecule 12:
        group = u.atoms[39:47]
        # select the rest of the universe's atoms:
        rest = u.atoms[:39]
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        # duplicate the group:
        group += group
        # store original positions of the rest:
        orig_rest_pos = rest.positions
        # get the expected result with duplicates:
        if compound == "group":
            # reference positions of UnWrapUniverse are known to be incorrect
            # for compound='group' if the group is not the entire system, so we
            # have to correct for that:
            ref_wrapped_pos = u.wrapped_coords("segments", center)[39:47]
        else:
            ref_wrapped_pos = u.wrapped_coords(compound, center)[39:47]
        ref_wrapped_pos = np.vstack((ref_wrapped_pos, ref_wrapped_pos))
        # wrap:
        group.wrap(compound=compound, center=center, inplace=True)
        # check for correct result:
        assert_almost_equal(
            group.atoms.positions, ref_wrapped_pos, decimal=self.precision
        )
        # check that the rest of the atoms are kept unmodified:
        assert_array_equal(rest.positions, orig_rest_pos)

    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("is_triclinic", (False, True))
    def test_wrap_com_cog_difference(self, compound, is_triclinic):
        # get a pristine test universe:
        u = UnWrapUniverse(is_triclinic=is_triclinic)
        # select molecule 5:
        group = u.atoms[6:9]
        # make first atom of molecule 5 much more heavy than the other two.
        # That way, the whole molecule's center of geometry will still lie
        # inside the first unit cell but its center of mass will lie outside
        # the first unit cell in negative x-direction.
        group.masses = [100.0, 1.0, 1.0]
        # wrap with center='cog':
        wrapped_pos_cog = group.wrap(
            compound=compound, center="cog", inplace=False
        )
        # get expected result:
        ref_wrapped_pos = u.wrapped_coords(compound, "cog")[6:9]
        # check for correctness:
        assert_almost_equal(
            wrapped_pos_cog, ref_wrapped_pos, decimal=self.precision
        )
        # wrap with center='com':
        wrapped_pos_com = group.wrap(
            compound=compound, center="com", inplace=False
        )
        # assert that the com result is shifted with respect to the cog result
        # by one box length in the x-direction:
        shift = np.array([10.0, 0.0, 0.0], dtype=np.float32)
        if compound == "atoms":
            # center argument must be ignored for compound='atoms':
            shift[0] = 0.0
        assert_almost_equal(
            wrapped_pos_cog, wrapped_pos_com - shift, decimal=self.precision
        )

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    def test_wrap_zero_mass_exception_safety(self, level, compound):
        # get a pristine test universe:
        u = UnWrapUniverse()
        # set masses of molecule 12 to zero:
        u.atoms[39:47].masses = 0.0
        # select group appropriate for compound:
        if compound == "group":
            group = u.atoms[39:47]  # molecule 12
        else:
            group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        # store original positions:
        orig_pos = group.atoms.positions
        if compound == "atoms":
            # wrap() must not care about masses if compound == 'atoms':
            group.wrap(compound=compound, center="com", inplace=True)
            ref_wrapped_pos = u.wrapped_coords(compound, "com")
            assert_almost_equal(
                group.atoms.positions, ref_wrapped_pos, decimal=self.precision
            )
        else:
            # try to wrap:
            with pytest.raises(ValueError):
                group.wrap(compound=compound, center="com", inplace=True)
            # make sure atom positions are unchanged:
            assert_array_equal(group.atoms.positions, orig_pos)

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    def test_wrap_wrong_center_exception_safety(self, level, compound):
        # get a pristine test universe:
        u = UnWrapUniverse()
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        # store original positions:
        orig_pos = group.atoms.positions
        if compound == "atoms":
            # wrap() must ignore the center argument if compound == 'atoms':
            group.wrap(compound=compound, center="com", inplace=True)
            ref_wrapped_pos = u.wrapped_coords(compound, "com")
            assert_almost_equal(
                group.atoms.positions, ref_wrapped_pos, decimal=self.precision
            )
        else:
            # try to wrap:
            with pytest.raises(ValueError):
                group.wrap(compound=compound, center="wrong", inplace=True)
            # make sure atom positions are unchanged:
            assert_array_equal(group.atoms.positions, orig_pos)

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_wrong_compound_exception_safety(self, level, center):
        # get a pristine test universe:
        u = UnWrapUniverse()
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        # store original positions:
        orig_pos = group.atoms.positions
        # try to wrap:
        with pytest.raises(ValueError):
            group.wrap(compound="wrong", center=center, inplace=True)
        # make sure atom positions are unchanged:
        assert_array_equal(group.atoms.positions, orig_pos)

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    def test_unwrap_no_masses_exception_safety(self, level, compound):
        # universe without masses:
        u = UnWrapUniverse(have_masses=False)
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        if compound == "atoms":
            # wrap() must not care about mass presence if compound == 'atoms':
            group.wrap(compound=compound, center="com", inplace=True)
            ref_wrapped_pos = u.wrapped_coords(compound, "com")
            assert_almost_equal(
                group.atoms.positions, ref_wrapped_pos, decimal=self.precision
            )
        else:
            # store original positions:
            orig_pos = group.atoms.positions
            # try to wrap:
            with pytest.raises(NoDataError):
                group.wrap(compound=compound, center="com", inplace=True)
            # make sure atom positions are unchanged:
            assert_array_equal(group.atoms.positions, orig_pos)

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_no_bonds_exception_safety(self, level, compound, center):
        # universe without bonds:
        u = UnWrapUniverse(have_bonds=False)
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        if compound == "fragments":
            # store original positions:
            orig_pos = group.atoms.positions
            # must raise an exception for fragments
            with pytest.raises(NoDataError):
                group.wrap(compound=compound, center=center, inplace=True)
            # make sure atom positions are unchanged:
            assert_array_equal(group.atoms.positions, orig_pos)
        else:
            # must not care about bonds if compound != 'fragments'
            group.wrap(compound=compound, center=center, inplace=True)
            ref_wrapped_pos = u.wrapped_coords(compound, center)
            assert_almost_equal(
                group.atoms.positions, ref_wrapped_pos, decimal=self.precision
            )

    @pytest.mark.parametrize("level", ("atoms", "residues", "segments"))
    @pytest.mark.parametrize(
        "compound",
        ("atoms", "group", "segments", "residues", "molecules", "fragments"),
    )
    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_no_molnums_exception_safety(self, level, compound, center):
        # universe without bonds:
        u = UnWrapUniverse(have_molnums=False)
        group = u.atoms
        # select topology level:
        if level == "residues":
            group = group.residues
        elif level == "segments":
            group = group.segments
        if compound == "molecules":
            # store original positions:
            orig_pos = group.atoms.positions
            # must raise an exception for molecules
            with pytest.raises(NoDataError):
                group.wrap(compound=compound, center=center, inplace=True)
            # make sure atom positions are unchanged:
            assert_array_equal(group.atoms.positions, orig_pos)
        else:
            # must not care about molnums if compound != 'molecules'
            group.wrap(compound=compound, center=center, inplace=True)
            ref_wrapped_pos = u.wrapped_coords(compound, center)
            assert_almost_equal(
                group.atoms.positions, ref_wrapped_pos, decimal=self.precision
            )


class TestWrapTRZ(object):
    """Tests the functionality of AtomGroup.wrap() using a TRZ universe."""

    precision = 5

    @pytest.fixture()
    def u(self):
        return mda.Universe(TRZ_psf, TRZ)

    @pytest.mark.parametrize("box", (np.array([1, 1]), np.zeros(6)))
    def test_wrap_box_fail(self, u, box):
        ag = u.atoms
        with pytest.raises(ValueError):
            ag.wrap(box=box)

    def test_wrap_small_box(self, u):
        ag = u.atoms[:100]
        box = u.dimensions
        box[:3] *= 0.5
        ag.wrap(box=box)
        assert_in_box(ag.positions, box)

    def test_wrap_large_box(self, u):
        ag = u.atoms[:100]
        box = u.dimensions
        box[:3] *= 2.0
        ag.wrap(box=box)
        assert_in_box(ag.positions, box)

    def test_wrap_atoms(self, u):
        ag = u.atoms[100:200]
        rescom = ag.wrap(compound="atoms", center="com", inplace=False)
        assert_in_box(rescom, u.dimensions)
        rescog = ag.wrap(compound="atoms", center="cog", inplace=False)
        assert_almost_equal(rescom, rescog, decimal=self.precision)

    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_group(self, u, center):
        ag = u.atoms[:100]
        ag.wrap(compound="group", center=center)
        if center == "com":
            ctrpos = ag.center_of_mass(wrap=False, compound="group")
        elif center == "cog":
            ctrpos = ag.center_of_geometry(wrap=False, compound="group")
        assert_in_box(ctrpos, u.dimensions)

    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_residues(self, u, center):
        ag = u.atoms[300:400].residues
        ag.wrap(compound="residues", center=center)
        if center == "com":
            ctrpos = ag.center_of_mass(wrap=False, compound="residues")
        elif center == "cog":
            ctrpos = ag.center_of_geometry(wrap=False, compound="residues")
        assert_in_box(ctrpos, u.dimensions)

    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_segments(self, u, center):
        ag = u.atoms[1000:1200]
        ag.wrap(compound="segments", center=center)
        if center == "com":
            ctrpos = ag.center_of_mass(wrap=False, compound="segments")
        elif center == "cog":
            ctrpos = ag.center_of_geometry(wrap=False, compound="segments")
        assert_in_box(ctrpos, u.dimensions)

    @pytest.mark.parametrize("center", ("com", "cog"))
    def test_wrap_fragments(self, u, center):
        ag = u.atoms[:250]
        ag.wrap(compound="fragments", center=center)
        if center == "com":
            ctrpos = ag.center_of_mass(wrap=False, compound="fragments")
        elif center == "cog":
            ctrpos = ag.center_of_geometry(wrap=False, compound="fragments")
        assert_in_box(ctrpos, u.dimensions)
